"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

plotting.py
~~~~~~~~~~~

Functionality to plot model data.

"""

import os
import re
import warnings

import plotly.offline as pltly
import jinja2

from calliope.exceptions import warn
from calliope.postprocess.plotting.capacity import plot_capacity
from calliope.postprocess.plotting.timeseries import plot_timeseries
from calliope.postprocess.plotting.transmission import plot_transmission
from calliope.postprocess.plotting.flows import plot_flows
from calliope.postprocess.plotting.util import type_of_script


def plot_summary(model, to_file=False, mapbox_access_token=None):
    """
    Plot a summary containing timeseries, installed capacities, and
    transmission plots. Returns a HTML string by default, returns None if
    ``to_file`` given (and saves the HTML string to file).

    Parameters
    ----------
    to_file : str, optional
        Path to output file to save HTML to.
    mapbox_access_token : str, optional
        (passed to plot_transmission) If given and a valid Mapbox API
        key, a Mapbox map is drawn for lat-lon coordinates, else
        (by default), a more simple built-in map.

    """
    subset = {"costs": ["monetary"]}

    timeseries = _plot(*plot_timeseries(model, subset=subset), html_only=True)
    capacity = _plot(*plot_capacity(model, subset=subset), html_only=True)
    if "loc_coordinates" in model._model_data:
        transmission = _plot(
            *plot_transmission(
                model, html_only=True, mapbox_access_token=mapbox_access_token
            ),
            html_only=True,
        )
    else:
        transmission = "<br><br><p>No location coordinates defined -<br>not plotting transmission.</p>"

    template_path = os.path.join(
        os.path.dirname(__file__), "..", "..", "config", "plots_template.html"
    )
    with open(template_path, "r") as f:
        html_template = jinja2.Template(f.read())

    if "solution_time" in model._model_data.attrs:
        solution_time = model._model_data.attrs["solution_time"] / 60
        time_finished = model._model_data.attrs["time_finished"]
        result_stats = "taking {:.2f} minutes to solve, completed at {}".format(
            solution_time, time_finished
        )
    else:
        result_stats = "inputs only"

    html = html_template.render(
        model_name=model.model_config["name"],
        calliope_version=model._model_data.attrs["calliope_version"],
        result_stats=result_stats,
        top=timeseries,
        bottom_left=capacity,
        bottom_right=transmission,
    )

    # Strip plotly-inserted style="..." attributes
    html = re.sub(r'style=".+?"', "", html)

    if to_file:
        with open(to_file, "w") as f:
            f.write(html)
    else:
        return html


def _plot(
    data,
    layout,
    html_only=False,
    to_file=False,
    layout_updates=None,
    plotly_kwarg_updates=None,
    # kwargs are included as they get passed through from the
    # plotting accessor method, but are not actually used
    **kwargs,
):

    # We will bee moving plotting out of Calliope core code (and a method of the model object)
    # in 0.7.0; current implementation is now untested.
    warnings.warn(
        "Plotting will no longer be available as a method of the Calliope model object in"
        "future versions of Calliope. In the meantime, as of v0.6.6, plotting is untested; this functionality should now be used with caution. We expect to reintroduce it as a seperate module in v0.7.0.",
        FutureWarning,
    )

    plotly_kwargs = dict(
        show_link=False,
        config={
            "displaylogo": False,
            "modeBarButtonsToRemove": ["sendDataToCloud"],
        },
    )

    if type_of_script() == "jupyter":
        plotter = pltly.iplot
        plotly_filename_key = "filename"
        pltly.init_notebook_mode(connected=False)
    else:
        plotter = pltly.plot
        plotly_filename_key = "image_filename"
        plotly_kwargs["auto_open"] = True
        plotly_kwargs["filename"] = "temp_plot.html"

    if layout_updates:
        layout = {**layout, **layout_updates}

    if plotly_kwarg_updates:
        plotly_kwargs = {**plotly_kwargs, **plotly_kwarg_updates}

    if to_file:
        filename, image_type = to_file.rsplit(".", 1)
        # Plotly can only save certain file types
        if image_type not in ["png", "jpeg", "svg", "webp"]:
            raise TypeError(
                "Unable to save plot as `{}`, choose from "
                "[`png`, `jpeg`, `svg`, `webp`]".format(image_type)
            )
        if "updatemenus" in layout:
            raise ValueError(
                "Unable to save multiple arrays to SVG, pick one array only"
            )
        else:
            plotly_kwargs.update(image=image_type)
            plotly_kwargs[plotly_filename_key] = filename

    if data:
        if html_only:
            return pltly.plot(
                {"data": data, "layout": layout},
                include_plotlyjs=False,
                output_type="div",
                **plotly_kwargs,
            )
        else:
            plotter({"data": data, "layout": layout}, **plotly_kwargs)
    else:
        raise ValueError("No data to plot.")


class ModelPlotMethods:
    def __init__(self, model):
        self._model = model

    _docstring_additions = """
    html_only : bool, optional; default = False
        Returns a html string for embedding the plot in a webpage
    to_file : False or str, optional; default = False
        Will save plot to file with the given name and extension.
        `to_file='plot.svg'` to save to SVG, `to_file='plot.png'` for
        a static PNG image.
        Allowed file extensions are: ['png', 'jpeg', 'svg', 'webp'].
    layout_updates : dict, optional
        The given dict will be merged with the Plotly layout dict
        generated by the Calliope plotting function, overwriting keys
        that already exist.
    plotly_kwarg_updates : dict, optional
        The given dict will be merged with the Plotly plot function's
        keyword arguments generated by the Calliope plotting function,
        overwriting keys that already exist.

    """

    def check_optimality(self):
        termination = self._model._model_data.attrs.get(
            "termination_condition", "did_not_yet_run"
        )
        # a MILP model which optimises to within the MIP gap, but does not fully
        # converge on the LP relaxation, may return as 'feasible', not 'optimal'
        if termination not in ["optimal", "did_not_yet_run", "feasible"]:
            warn("Model termination condition was not optimal. Plotting may fail!")

    def timeseries(self, **kwargs):
        self.check_optimality()
        data, layout = plot_timeseries(self._model, **kwargs)
        return _plot(data, layout, **kwargs)

    timeseries.__doc__ = plot_timeseries.__doc__.rstrip() + _docstring_additions

    def capacity(self, **kwargs):
        self.check_optimality()
        data, layout = plot_capacity(self._model, **kwargs)
        return _plot(data, layout, **kwargs)

    capacity.__doc__ = plot_capacity.__doc__.rstrip() + _docstring_additions

    def transmission(self, **kwargs):
        self.check_optimality()
        data, layout = plot_transmission(self._model, **kwargs)
        return _plot(data, layout, **kwargs)

    transmission.__doc__ = plot_transmission.__doc__.rstrip() + _docstring_additions

    def flows(self, **kwargs):
        self.check_optimality()
        data, layout = plot_flows(self._model, **kwargs)
        return _plot(data, layout, **kwargs)

    flows.__doc__ = plot_flows.__doc__.rstrip() + _docstring_additions

    def summary(self, **kwargs):
        self.check_optimality()
        return plot_summary(self._model, **kwargs)

    summary.__doc__ = plot_summary.__doc__



import pytest
from pytest import approx
import pandas as pd

import calliope
from calliope.test.common.util import build_test_model as build_model
from calliope.test.common.util import check_error_or_warning


class TestNationalScaleExampleModelSenseChecks:
    def test_group_prod_min(self):
        model = calliope.examples.national_scale(
            scenario="cold_fusion_with_production_share"
        )
        model.run()

        df_carrier_prod = (
            model.get_formatted_array("carrier_prod")
            .loc[dict(carriers="power")]
            .sum("locs")
            .sum("timesteps")
            .to_pandas()
        )

        prod_share = (
            df_carrier_prod.loc[["cold_fusion", "csp"]].sum()
            / df_carrier_prod.loc[["ccgt", "cold_fusion", "csp"]].sum()
        )

        assert prod_share == approx(0.85)

    def test_group_cap_max(self):
        model = calliope.examples.national_scale(
            scenario="cold_fusion_with_capacity_share"
        )
        model.run()

        cap_share = (
            model.get_formatted_array("energy_cap")
            .loc[{"techs": ["cold_fusion", "csp"]}]
            .sum()
            / model.get_formatted_array("energy_cap")
            .loc[{"techs": ["ccgt", "cold_fusion", "csp"]}]
            .sum()
        )

        assert cap_share == approx(0.2)

    def test_systemwide_equals(self):
        model = calliope.examples.national_scale(
            override_dict={
                "techs.ccgt.constraints.energy_cap_max_systemwide": 10000,
                "techs.ac_transmission.constraints.energy_cap_equals_systemwide": 6000,
            }
        )
        model.run()
        # Check that setting `_equals` to a finite value leads to forcing
        assert (
            model.get_formatted_array("energy_cap").loc[{"techs": "ccgt"}].sum()
            == 10000
        )
        assert (
            model.get_formatted_array("energy_cap")
            .loc[{"techs": "ac_transmission:region1"}]
            .sum()
            == 6000
        )

    def test_reserve_margin(self):
        model = calliope.examples.national_scale(scenario="reserve_margin")

        model.run()

        # constraint_string = '-Inf : -1.1 * ( carrier_con[region1::demand_power::power,2005-01-05 16:00:00] + carrier_con[region2::demand_power::power,2005-01-05 16:00:00] ) / timestep_resolution[2005-01-05 16:00:00] - energy_cap[region1::ccgt] - energy_cap[region1-3::csp] - energy_cap[region1-2::csp] - energy_cap[region1-1::csp] :   0.0'

        # FIXME: capture Pyomo's print output...
        # assert constraint_string in model._backend_model.reserve_margin_constraint.pprint()

        assert float(model.results.cost.sum()) == approx(282487.35489)


@pytest.mark.filterwarnings("ignore:(?s).*Integer:calliope.exceptions.ModelWarning")
class TestUrbanScaleMILP:
    def test_asynchronous_prod_con(self):
        def _get_prod_con(model, prod_con):
            return (
                model.get_formatted_array("carrier_{}".format(prod_con))
                .loc[{"techs": "heat_pipes:X1", "carriers": "heat"}]
                .to_pandas()
                .dropna(how="all")
            )

        m = calliope.examples.urban_scale(override_dict={"run.zero_threshold": 1e-6})
        m.run()
        _prod = _get_prod_con(m, "prod")
        _con = _get_prod_con(m, "con")
        assert any(((_con < 0) & (_prod > 0)).any()) is True

        m_bin = calliope.examples.urban_scale(
            override_dict={
                "techs.heat_pipes.constraints.force_asynchronous_prod_con": True,
                "run.solver_options.mipgap": 0.05,
                "run.zero_threshold": 1e-6,
            }
        )
        m_bin.run()
        _prod = _get_prod_con(m_bin, "prod")
        _con = _get_prod_con(m_bin, "con")
        assert any(((_con < 0) & (_prod > 0)).any()) is False


class TestModelSettings:
    def model(self, feasibility, cap_val):
        override_dict = {
            "locations.0.techs": {"test_supply_elec": {}, "test_demand_elec": {}},
            "links.0,1.exists": False,
            # pick a time subset where demand is uniformally -10 throughout
            "model.subset_time": ["2005-01-01 06:00:00", "2005-01-01 08:00:00"],
            "run.ensure_feasibility": feasibility,
            "run.bigM": 1e3,
            # Allow setting resource and energy_cap_max/equals to force infeasibility
            "techs.test_supply_elec.constraints": {
                "resource": cap_val,
                "energy_eff": 1,
                "energy_cap_equals": 15,
                "force_resource": True,
            },
        }
        model = build_model(override_dict=override_dict, scenario="investment_costs")
        model.run()

        return model

    @pytest.fixture
    def model_no_unmet(self):
        return self.model(True, 10)

    @pytest.fixture
    def model_unmet_demand(self):
        return self.model(True, 5)

    @pytest.fixture
    def model_unused_supply(self):
        return self.model(True, 15)

    def test_unmet_demand_zero(self, model_no_unmet):

        # Feasible case, but unmet_demand/unused_supply is not deleted
        for i in ["unmet_demand", "unused_supply"]:
            assert hasattr(model_no_unmet._backend_model, i)
        assert "unmet_demand" in model_no_unmet._model_data.data_vars.keys()
        assert "unused_supply" not in model_no_unmet._model_data.data_vars.keys()
        assert (model_no_unmet._model_data["unmet_demand"] == 0).all()

    def test_unmet_demand_nonzero(self, model_unmet_demand):

        # Infeasible case, unmet_demand is required
        assert hasattr(model_unmet_demand._backend_model, "unmet_demand")
        assert hasattr(model_unmet_demand._backend_model, "unused_supply")
        assert model_unmet_demand._model_data["unmet_demand"].sum() == 15
        assert "unused_supply" not in model_unmet_demand._model_data.data_vars.keys()

    def test_unmet_supply_nonzero(self, model_unused_supply):
        # Infeasible case, unused_supply is required
        assert hasattr(model_unused_supply._backend_model, "unmet_demand")
        assert hasattr(model_unused_supply._backend_model, "unused_supply")
        assert model_unused_supply._model_data["unmet_demand"].sum() == -15
        assert "unused_supply" not in model_unused_supply._model_data.data_vars.keys()

    def test_expected_impact_on_objective_function_value(
        self, model_no_unmet, model_unmet_demand, model_unused_supply
    ):
        assert (
            model_unused_supply._backend_model.obj.expr()
            - model_no_unmet._backend_model.obj.expr()
            == approx(1e3 * 15)
        )

        assert (
            model_unmet_demand._backend_model.obj.expr()
            - model_no_unmet._backend_model.obj.expr()
            == approx(1e3 * 15)
        )

    @pytest.mark.parametrize("override", (5, 15))
    def test_expected_infeasible_result(self, override):
        model = self.model(False, override)

        assert not hasattr(model._backend_model, "unmet_demand")
        assert not hasattr(model._backend_model, "unused_supply")
        assert not model._model_data.attrs["termination_condition"] == "optimal"


class TestEnergyCapacityPerStorageCapacity:
    @pytest.fixture
    def model_file(self):
        return "energy_cap_per_storage_cap.yaml"

    @pytest.mark.filterwarnings(
        "ignore:(?s).*`energy_cap_per_storage_cap_min/max/equals`:calliope.exceptions.ModelWarning"
    )
    def test_no_constraint_set(self, model_file):
        model = build_model(model_file=model_file)
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = (
            model.get_formatted_array("energy_cap")
            .loc[{"techs": "my_storage"}]
            .sum()
            .item()
        )
        storage_capacity = (
            model.get_formatted_array("storage_cap")
            .loc[{"techs": "my_storage"}]
            .sum()
            .item()
        )
        assert energy_capacity == pytest.approx(10)
        assert storage_capacity == pytest.approx(180)
        assert storage_capacity != pytest.approx(1 / 10 * energy_capacity)

    def test_equals(self, model_file):
        model = build_model(model_file=model_file, scenario="equals")
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = (
            model.get_formatted_array("energy_cap")
            .loc[{"techs": "my_storage"}]
            .sum()
            .item()
        )
        storage_capacity = (
            model.get_formatted_array("storage_cap")
            .loc[{"techs": "my_storage"}]
            .sum()
            .item()
        )
        assert storage_capacity == pytest.approx(1 / 10 * energy_capacity)

    def test_max(self, model_file):
        model = build_model(model_file=model_file, scenario="max")
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = (
            model.get_formatted_array("energy_cap")
            .loc[{"techs": "my_storage"}]
            .sum()
            .item()
        )
        storage_capacity = (
            model.get_formatted_array("storage_cap")
            .loc[{"techs": "my_storage"}]
            .sum()
            .item()
        )
        assert energy_capacity == pytest.approx(10)
        assert storage_capacity == pytest.approx(1000)

    def test_min(self, model_file):
        model = build_model(model_file=model_file, scenario="min")
        model.run()
        assert model.results.termination_condition == "optimal"
        energy_capacity = (
            model.get_formatted_array("energy_cap")
            .loc[{"techs": "my_storage"}]
            .sum()
            .item()
        )
        storage_capacity = (
            model.get_formatted_array("storage_cap")
            .loc[{"techs": "my_storage"}]
            .sum()
            .item()
        )
        assert energy_capacity == pytest.approx(180)
        assert storage_capacity == pytest.approx(180)

    def test_operate_mode(self, model_file):
        model = build_model(model_file=model_file, scenario="operate_mode_min")
        model.run()

    @pytest.mark.parametrize(
        "horizon_window", [(24, 24), (48, 48), (72, 48), (144, 24)]
    )
    def test_operate_mode_horizon_window(self, model_file, horizon_window):
        horizon, window = horizon_window
        override_dict = {
            "model.subset_time": ["2005-01-01", "2005-01-05"],
            "run.operation.horizon": horizon,
            "run.operation.window": window,
        }
        model = build_model(
            model_file=model_file,
            scenario="operate_mode_min",
            override_dict=override_dict,
        )
        model.run()



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

Export.py
~~~~~~~~~~~~~~~~~

Energy export constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import (
    get_param,
    get_loc_tech_carriers,
    get_loc_tech,
    loc_tech_is_in,
)

ORDER = 30  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]

    if "loc_carriers_update_system_balance_constraint" in sets:
        for loc_carrier, timestep in (
            backend_model.loc_carriers_update_system_balance_constraint
            * backend_model.timesteps
        ):
            update_system_balance_constraint(backend_model, loc_carrier, timestep)

    if "loc_tech_carriers_export_balance_constraint" in sets:
        backend_model.export_balance_constraint = po.Constraint(
            backend_model.loc_tech_carriers_export_balance_constraint,
            backend_model.timesteps,
            rule=export_balance_constraint_rule,
        )

    if "loc_techs_update_costs_var_constraint" in sets:
        for cost, loc_tech, timestep in (
            backend_model.costs
            * backend_model.loc_techs_update_costs_var_constraint
            * backend_model.timesteps
        ):
            update_costs_var_constraint(backend_model, cost, loc_tech, timestep)

    if "loc_tech_carriers_export_max_constraint" in sets:
        backend_model.export_max_constraint = po.Constraint(
            backend_model.loc_tech_carriers_export_max_constraint,
            backend_model.timesteps,
            rule=export_max_constraint_rule,
        )


def update_system_balance_constraint(backend_model, loc_carrier, timestep):
    """
    Update system balance constraint (from energy_balance.py) to include export

    Math given in :func:`~calliope.backend.pyomo.constraints.energy_balance.system_balance_constraint_rule`
    """
    prod, con, export = get_loc_tech_carriers(backend_model, loc_carrier)

    backend_model.system_balance[loc_carrier, timestep].expr += -1 * (
        sum(
            backend_model.carrier_export[loc_tech_carrier, timestep]
            for loc_tech_carrier in export
        )
    )

    return None


def export_balance_constraint_rule(backend_model, loc_tech_carrier, timestep):
    """
    Ensure no technology can 'pass' its export capability to another technology
    with the same carrier_out, by limiting its export to the capacity of its production


    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\geq \\boldsymbol{carrier_{export}}(loc::tech::carrier, timestep)
            \\quad \\forall loc::tech::carrier \\in locs::tech::carriers_{export},
            \\forall timestep \\in timesteps
    """

    return (
        backend_model.carrier_prod[loc_tech_carrier, timestep]
        >= backend_model.carrier_export[loc_tech_carrier, timestep]
    )


def update_costs_var_constraint(backend_model, cost, loc_tech, timestep):
    """
    Update time varying cost constraint (from costs.py) to include export

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost_{var}}(cost, loc::tech, timestep) +=
            cost_{export}(cost, loc::tech, timestep) \\times
            \\boldsymbol{carrier_{export}}(loc::tech::carrier, timestep)
            * timestep_{weight} \\quad \\forall cost \\in costs,
            \\forall loc::tech \\in loc::techs_{cost_{var}, export},
            \\forall timestep \\in timesteps

    """
    model_data_dict = backend_model.__calliope_model_data["data"]

    loc_tech_carrier = model_data_dict["lookup_loc_techs_export"][(loc_tech)]
    weight = backend_model.timestep_weights[timestep]

    cost_export = (
        get_param(backend_model, "cost_export", (cost, loc_tech, timestep))
        * backend_model.carrier_export[loc_tech_carrier, timestep]
        * weight
    )

    backend_model.cost_var_rhs[cost, loc_tech, timestep].expr += cost_export


def export_max_constraint_rule(backend_model, loc_tech_carrier, timestep):
    """
    Set maximum export. All exporting technologies.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{export}}(loc::tech::carrier, timestep)
            \\leq export_{cap}(loc::tech)
            \\quad \\forall loc::tech::carrier \\in locs::tech::carriers_{export},
            \\forall timestep \\in timesteps

    If the technology is defined by integer units, not a continuous capacity,
    this constraint becomes:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{export}}(loc::tech::carrier, timestep)
            \\leq export_{cap}(loc::tech) \\times
            \\boldsymbol{operating_{units}}(loc::tech, timestep)

    """

    loc_tech = get_loc_tech(loc_tech_carrier)

    if loc_tech_is_in(backend_model, loc_tech, "loc_techs_milp"):
        operating_units = backend_model.operating_units[loc_tech, timestep]
    else:
        operating_units = 1

    export_cap = get_param(backend_model, "export_cap", loc_tech)
    return (
        backend_model.carrier_export[loc_tech_carrier, timestep]
        <= export_cap * operating_units
    )



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

energy_balance.py
~~~~~~~~~~~~~~~~~

Energy balance constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import (
    get_param,
    get_previous_timestep,
    get_loc_tech_carriers,
    loc_tech_is_in,
)

ORDER = 10  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]
    run_config = backend_model.__calliope_run_config

    if "loc_carriers_system_balance_constraint" in sets:
        backend_model.system_balance = po.Expression(
            backend_model.loc_carriers_system_balance_constraint,
            backend_model.timesteps,
            initialize=0.0,
        )

        backend_model.system_balance_constraint = po.Constraint(
            backend_model.loc_carriers_system_balance_constraint,
            backend_model.timesteps,
            rule=system_balance_constraint_rule,
        )

    if "loc_techs_balance_supply_constraint" in sets:
        backend_model.balance_supply_constraint = po.Constraint(
            backend_model.loc_techs_balance_supply_constraint,
            backend_model.timesteps,
            rule=balance_supply_constraint_rule,
        )

    if "loc_techs_balance_demand_constraint" in sets:
        # Add a required_resource expression to share computed values between constraints
        backend_model.required_resource = po.Expression(
            backend_model.loc_techs_balance_demand_constraint,
            backend_model.timesteps,
            initialize=0.0,
        )
        backend_model.balance_demand_constraint = po.Constraint(
            backend_model.loc_techs_balance_demand_constraint,
            backend_model.timesteps,
            rule=balance_demand_constraint_rule,
        )

    if "loc_techs_balance_transmission_constraint" in sets:
        backend_model.balance_transmission_constraint = po.Constraint(
            backend_model.loc_techs_balance_transmission_constraint,
            backend_model.timesteps,
            rule=balance_transmission_constraint_rule,
        )

    if "loc_techs_resource_availability_supply_plus_constraint" in sets:
        backend_model.balance_supply_plus_constraint = po.Constraint(
            backend_model.loc_techs_resource_availability_supply_plus_constraint,
            backend_model.timesteps,
            rule=balance_supply_plus_constraint_rule,
        )

    if "loc_techs_balance_supply_plus_constraint" in sets:
        backend_model.resource_availability_supply_plus_constraint = po.Constraint(
            backend_model.loc_techs_balance_supply_plus_constraint,
            backend_model.timesteps,
            rule=resource_availability_supply_plus_constraint_rule,
        )

    if "loc_techs_balance_storage_constraint" in sets:
        backend_model.balance_storage_constraint = po.Constraint(
            backend_model.loc_techs_balance_storage_constraint,
            backend_model.timesteps,
            rule=balance_storage_constraint_rule,
        )

    if "loc_techs_balance_storage_inter_cluster_constraint" in sets:
        backend_model.balance_storage_inter_cluster_constraint = po.Constraint(
            backend_model.loc_techs_balance_storage_inter_cluster_constraint,
            backend_model.datesteps,
            rule=balance_storage_inter_cluster_rule,
        )

    if (
        "loc_techs_storage_initial_constraint" in sets
        and run_config["mode"] != "operate"
    ):
        backend_model.storage_initial_constraint = po.Constraint(
            backend_model.loc_techs_storage_initial_constraint,
            rule=storage_initial_rule,
        )


def system_balance_constraint_rule(backend_model, loc_carrier, timestep):
    """
    System balance ensures that, within each location, the production and
    consumption of each carrier is balanced.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier_{prod} \\in loc::carrier} \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            + \\sum_{loc::tech::carrier_{con} \\in loc::carrier} \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            + \\sum_{loc::tech::carrier_{export} \\in loc::carrier} \\boldsymbol{carrier_{export}}(loc::tech::carrier, timestep)
            \\quad \\forall loc::carrier \\in loc::carriers, \\forall timestep \\in timesteps

    """
    prod, con, export = get_loc_tech_carriers(backend_model, loc_carrier)
    if backend_model.__calliope_run_config.get("ensure_feasibility", False):
        unmet_demand = backend_model.unmet_demand[loc_carrier, timestep]
        unused_supply = backend_model.unused_supply[loc_carrier, timestep]
    else:
        unmet_demand = unused_supply = 0

    backend_model.system_balance[loc_carrier, timestep].expr = (
        sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in prod
        )
        + sum(
            backend_model.carrier_con[loc_tech_carrier, timestep]
            for loc_tech_carrier in con
        )
        + unmet_demand
        + unused_supply
    )

    return backend_model.system_balance[loc_carrier, timestep] == 0


def balance_supply_constraint_rule(backend_model, loc_tech, timestep):
    """
    Limit production from supply techs to their available resource

    .. container:: scrolling-wrapper

        .. math::

            min\\_use(loc::tech) \\times available\\_resource(loc::tech, timestep) \\leq
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep)}
            \\geq available\\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{supply}, \\forall timestep \\in timesteps

    If :math:`force\\_resource(loc::tech)` is set:

    .. container:: scrolling-wrapper

        .. math::

            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep)}
            = available\\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{supply}, \\forall timestep \\in timesteps

    Where:

    .. container:: scrolling-wrapper

        .. math::

            available\\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource\\_scale(loc::tech)

    if :math:`loc::tech` is in :math:`loc::techs_{area}`:

    .. container:: scrolling-wrapper

        .. math::

            available\\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource\\_scale(loc::tech) \\times \\boldsymbol{resource_{area}}(loc::tech)

    """
    model_data_dict = backend_model.__calliope_model_data["data"]

    resource = get_param(backend_model, "resource", (loc_tech, timestep))
    energy_eff = get_param(backend_model, "energy_eff", (loc_tech, timestep))
    resource_scale = get_param(backend_model, "resource_scale", loc_tech)
    force_resource = get_param(backend_model, "force_resource", loc_tech)
    loc_tech_carrier = model_data_dict["lookup_loc_techs"][loc_tech]
    min_use = get_param(backend_model, "resource_min_use", (loc_tech, timestep))
    resource_unit = get_param(backend_model, "resource_unit", loc_tech)

    if po.value(energy_eff) == 0:
        return backend_model.carrier_prod[loc_tech_carrier, timestep] == 0
    else:
        carrier_prod = (
            backend_model.carrier_prod[loc_tech_carrier, timestep] / energy_eff
        )

    if po.value(resource_unit) == "energy_per_area":
        available_resource = (
            resource * resource_scale * backend_model.resource_area[loc_tech]
        )
    elif po.value(resource_unit) == "energy_per_cap":
        available_resource = (
            resource * resource_scale * backend_model.energy_cap[loc_tech]
        )
    else:
        available_resource = resource * resource_scale

    if po.value(force_resource):
        return carrier_prod == available_resource
    elif min_use:
        return min_use * available_resource <= carrier_prod <= available_resource
    else:
        return carrier_prod <= available_resource


def balance_demand_constraint_rule(backend_model, loc_tech, timestep):
    """
    Limit consumption from demand techs to their required resource.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep) \\times \\eta_{energy}(loc::tech, timestep) \\geq
            required\\_resource(loc::tech, timestep) \\quad \\forall loc::tech \\in loc::techs_{demand},
            \\forall timestep \\in timesteps

    If :math:`force\\_resource(loc::tech)` is set:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep) \\times \\eta_{energy}(loc::tech, timestep) =
            required\\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{demand}, \\forall timestep \\in timesteps

    Where:

    .. container:: scrolling-wrapper

        .. math::

            required\\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource\\_scale(loc::tech)

    if :math:`loc::tech` is in :math:`loc::techs_{area}`:

    .. container:: scrolling-wrapper

        .. math::

            required\\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource\\_scale(loc::tech) \\times \\boldsymbol{resource_{area}}(loc::tech)

    """
    model_data_dict = backend_model.__calliope_model_data["data"]

    resource = get_param(backend_model, "resource", (loc_tech, timestep))
    energy_eff = get_param(backend_model, "energy_eff", (loc_tech, timestep))
    resource_scale = get_param(backend_model, "resource_scale", loc_tech)
    force_resource = get_param(backend_model, "force_resource", loc_tech)
    resource_unit = get_param(backend_model, "resource_unit", loc_tech)

    loc_tech_carrier = model_data_dict["lookup_loc_techs"][loc_tech]
    carrier_con = backend_model.carrier_con[loc_tech_carrier, timestep] * energy_eff

    if po.value(resource_unit) == "energy_per_area":
        required_resource = (
            resource * resource_scale * backend_model.resource_area[loc_tech]
        )
    elif po.value(resource_unit) == "energy_per_cap":
        required_resource = (
            resource * resource_scale * backend_model.energy_cap[loc_tech]
        )
    else:
        required_resource = resource * resource_scale

    # We save the expression to the backend_model so it can be used elsewhere,
    # e.g. in the group constraints
    backend_model.required_resource[loc_tech, timestep] = required_resource

    if po.value(force_resource):
        return carrier_con == backend_model.required_resource[loc_tech, timestep]
    else:
        return carrier_con >= backend_model.required_resource[loc_tech, timestep]


def resource_availability_supply_plus_constraint_rule(
    backend_model, loc_tech, timestep
):
    """
    Limit production from supply_plus techs to their available resource.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{con}}(loc::tech, timestep)
            \\leq available\\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{supply^{+}}, \\forall timestep \\in timesteps

    If :math:`force\\_resource(loc::tech)` is set:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{con}}(loc::tech, timestep)
            = available\\_resource(loc::tech, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{supply^{+}}, \\forall timestep \\in timesteps

    Where:

    .. container:: scrolling-wrapper

        .. math::

            available\\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource_{scale}(loc::tech)

    if :math:`loc::tech` is in :math:`loc::techs_{area}`:

    .. container:: scrolling-wrapper

        .. math::

            available\\_resource(loc::tech, timestep) = resource(loc::tech, timestep)
            \\times resource_{scale}(loc::tech)
            \\times resource_{area}(loc::tech)

    """
    resource = get_param(backend_model, "resource", (loc_tech, timestep))
    resource_scale = get_param(backend_model, "resource_scale", loc_tech)
    force_resource = get_param(backend_model, "force_resource", loc_tech)
    resource_unit = get_param(backend_model, "resource_unit", loc_tech)

    if po.value(resource_unit) == "energy_per_area":
        available_resource = (
            resource * resource_scale * backend_model.resource_area[loc_tech]
        )
    elif po.value(resource_unit) == "energy_per_cap":
        available_resource = (
            resource * resource_scale * backend_model.energy_cap[loc_tech]
        )
    else:
        available_resource = resource * resource_scale

    if po.value(force_resource):
        return backend_model.resource_con[loc_tech, timestep] == available_resource
    else:
        return backend_model.resource_con[loc_tech, timestep] <= available_resource


def balance_transmission_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance carrier production and consumption of transmission technologies

    .. container:: scrolling-wrapper

        .. math::

            -1 * \\boldsymbol{carrier_{con}}(loc_{from}::tech:loc_{to}::carrier, timestep)
            \\times \\eta_{energy}(loc::tech, timestep)
            = \\boldsymbol{carrier_{prod}}(loc_{to}::tech:loc_{from}::carrier, timestep)
            \\quad \\forall loc::tech:loc \\in locs::techs:locs_{transmission},
            \\forall timestep \\in timesteps

    Where a link is the connection between :math:`loc_{from}::tech:loc_{to}`
    and :math:`loc_{to}::tech:loc_{from}` for locations `to` and `from`.

    """
    model_data_dict = backend_model.__calliope_model_data["data"]

    energy_eff = get_param(backend_model, "energy_eff", (loc_tech, timestep))
    loc_tech_carrier = model_data_dict["lookup_loc_techs"][loc_tech]
    remote_loc_tech = model_data_dict["lookup_remotes"][loc_tech]
    remote_loc_tech_carrier = model_data_dict["lookup_loc_techs"][remote_loc_tech]

    if (
        loc_tech_is_in(backend_model, remote_loc_tech, "loc_techs_transmission")
        and get_param(backend_model, "energy_prod", (loc_tech)).value == 1
    ):
        return (
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            == -1
            * backend_model.carrier_con[remote_loc_tech_carrier, timestep]
            * energy_eff
        )
    else:
        return po.Constraint.NoConstraint


def balance_supply_plus_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance carrier production and resource consumption of supply_plus technologies
    alongside any use of resource storage.

    .. _system_balance_constraint:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep) =
            \\boldsymbol{storage}(loc::tech, timestep_{previous})
            \\times (1 - storage\\_loss(loc::tech, timestep))^{timestep\\_resolution(timestep)} +
            \\boldsymbol{resource_{con}}(loc::tech, timestep) \\times \\eta_{resource}(loc::tech, timestep) -
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep) \\times \\eta_{parasitic}(loc::tech, timestep)}
            \\quad \\forall loc::tech \\in loc::techs_{supply^{+}}, \\forall timestep \\in timesteps

    If *no* storage is defined for the technology, this reduces to:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{con}}(loc::tech, timestep) \\times \\eta_{resource}(loc::tech, timestep) =
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep) \\times \\eta_{parasitic}(loc::tech, timestep)}
            \\quad \\forall loc::tech \\in loc::techs_{supply^{+}}, \\forall timestep \\in timesteps

    """

    model_data_dict = backend_model.__calliope_model_data["data"]
    run_config = backend_model.__calliope_run_config

    resource_eff = get_param(backend_model, "resource_eff", (loc_tech, timestep))
    energy_eff = get_param(backend_model, "energy_eff", (loc_tech, timestep))
    parasitic_eff = get_param(backend_model, "parasitic_eff", (loc_tech, timestep))
    total_eff = energy_eff * parasitic_eff

    if po.value(total_eff) == 0:
        carrier_prod = 0
    else:
        loc_tech_carrier = model_data_dict["lookup_loc_techs"][loc_tech]
        carrier_prod = (
            backend_model.carrier_prod[loc_tech_carrier, timestep] / total_eff
        )

    # A) Case where no storage allowed
    if not loc_tech_is_in(backend_model, loc_tech, "loc_techs_store"):
        return (
            backend_model.resource_con[loc_tech, timestep] * resource_eff
            == carrier_prod
        )

    # B) Case where storage is allowed
    else:
        resource = backend_model.resource_con[loc_tech, timestep] * resource_eff
        # Pyomo returns the order 1-indexed, but we want 0-indexing
        current_timestep = backend_model.timesteps.ord(timestep) - 1
        if current_timestep == 0 and not run_config["cyclic_storage"]:
            storage_previous_step = (
                get_param(backend_model, "storage_initial", loc_tech)
                * backend_model.storage_cap[loc_tech]
            )
        elif (
            hasattr(backend_model, "storage_inter_cluster")
            and model_data_dict["lookup_cluster_first_timestep"][timestep]
        ):
            storage_previous_step = 0
        else:
            if (
                hasattr(backend_model, "clusters")
                and model_data_dict["lookup_cluster_first_timestep"][timestep]
            ):
                previous_step = model_data_dict["lookup_cluster_last_timestep"][
                    timestep
                ]
            elif current_timestep == 0 and run_config["cyclic_storage"]:
                previous_step = backend_model.timesteps[-1]
            else:
                previous_step = get_previous_timestep(backend_model.timesteps, timestep)
            storage_loss = get_param(backend_model, "storage_loss", loc_tech)
            time_resolution = backend_model.timestep_resolution[previous_step]
            storage_previous_step = (
                (1 - storage_loss) ** time_resolution
            ) * backend_model.storage[loc_tech, previous_step]

        return (
            backend_model.storage[loc_tech, timestep]
            == storage_previous_step + resource - carrier_prod
        )


def balance_storage_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance carrier production and consumption of storage technologies,
    alongside any use of the stored volume.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep) =
            \\boldsymbol{storage}(loc::tech, timestep_{previous})
            \\times (1 - storage\\_loss(loc::tech, timestep))^{resolution(timestep)}
            - \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            \\times \\eta_{energy}(loc::tech, timestep)
            - \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{\\eta_{energy}(loc::tech, timestep)}
            \\quad \\forall loc::tech \\in loc::techs_{storage}, \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data["data"]
    run_config = backend_model.__calliope_run_config

    energy_eff = get_param(backend_model, "energy_eff", (loc_tech, timestep))

    if po.value(energy_eff) == 0:
        carrier_prod = 0
    else:
        loc_tech_carrier = model_data_dict["lookup_loc_techs"][loc_tech]
        carrier_prod = (
            backend_model.carrier_prod[loc_tech_carrier, timestep] / energy_eff
        )

    carrier_con = backend_model.carrier_con[loc_tech_carrier, timestep] * energy_eff

    # Pyomo returns the order 1-indexed, but we want 0-indexing
    current_timestep = backend_model.timesteps.ord(timestep) - 1
    if current_timestep == 0 and not run_config["cyclic_storage"]:
        storage_previous_step = (
            get_param(backend_model, "storage_initial", loc_tech)
            * backend_model.storage_cap[loc_tech]
        )
    elif (
        hasattr(backend_model, "storage_inter_cluster")
        and model_data_dict["lookup_cluster_first_timestep"][timestep]
    ):
        storage_previous_step = 0
    else:
        if (
            hasattr(backend_model, "clusters")
            and model_data_dict["lookup_cluster_first_timestep"][timestep]
        ):
            previous_step = model_data_dict["lookup_cluster_last_timestep"][timestep]
        elif current_timestep == 0 and run_config["cyclic_storage"]:
            previous_step = backend_model.timesteps[-1]
        else:
            previous_step = get_previous_timestep(backend_model.timesteps, timestep)
        storage_loss = get_param(backend_model, "storage_loss", loc_tech)
        time_resolution = backend_model.timestep_resolution[previous_step]
        storage_previous_step = (
            (1 - storage_loss) ** time_resolution
        ) * backend_model.storage[loc_tech, previous_step]

    return (
        backend_model.storage[loc_tech, timestep]
        == storage_previous_step - carrier_prod - carrier_con
    )


def balance_storage_inter_cluster_rule(backend_model, loc_tech, datestep):
    """
    When clustering days, to reduce the timeseries length, balance the daily stored
    energy across all days of the original timeseries.

    `Ref: DOI 10.1016/j.apenergy.2018.01.023 <https://doi.org/10.1016/j.apenergy.2018.01.023>`_

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{inter\\_cluster}}(loc::tech, datestep) =
            \\boldsymbol{storage_{inter\\_cluster}}(loc::tech, datestep_{previous})
            \\times (1 - storage\\_loss(loc::tech, timestep))^{24}
            + \\boldsymbol{storage}(loc::tech, timestep_{final, cluster(datestep))})
            \\quad \\forall loc::tech \\in loc::techs_{store}, \\forall datestep \\in datesteps

    Where :math:`timestep_{final, cluster(datestep_{previous}))}` is the final timestep of the
    cluster in the clustered timeseries corresponding to the previous day
    """
    run_config = backend_model.__calliope_run_config
    # Pyomo returns the order 1-indexed, but we want 0-indexing
    current_datestep = backend_model.datesteps.ord(datestep) - 1

    if current_datestep == 0 and not run_config["cyclic_storage"]:
        storage_previous_step = get_param(backend_model, "storage_initial", loc_tech)
        storage_intra = 0
    else:
        if current_datestep == 0 and run_config["cyclic_storage"]:
            previous_step = backend_model.datesteps[-1]
        else:
            previous_step = get_previous_timestep(backend_model.datesteps, datestep)
        storage_loss = get_param(backend_model, "storage_loss", loc_tech)
        storage_previous_step = (
            (1 - storage_loss) ** 24
        ) * backend_model.storage_inter_cluster[loc_tech, previous_step]
        final_timestep = backend_model.__calliope_model_data["data"][
            "lookup_datestep_last_cluster_timestep"
        ][previous_step]
        storage_intra = backend_model.storage[loc_tech, final_timestep]
    return (
        backend_model.storage_inter_cluster[loc_tech, datestep]
        == storage_previous_step + storage_intra
    )


def storage_initial_rule(backend_model, loc_tech):
    """
    If storage is cyclic, allow an initial storage to still be set. This is
    applied to the storage of the final timestep/datestep of the series as that,
    in cyclic storage, is the 'storage_previous_step' for the first
    timestep/datestep.

    If clustering and ``storage_inter_cluster`` exists:

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{inter\\_cluster}}(loc::tech, datestep_{final})
            \\times ((1 - storage_loss) ** 24) = storage_{initial}(loc::tech) \\times storage_{cap}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{store}, \\forall datestep \\in datesteps

    Where :math:`datestep_{final}` is the last datestep of the timeseries

    Else:
    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep_{final})
            \\times ((1 - storage_loss) ** 24) = storage_{initial}(loc::tech) \\times storage_{cap}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{store}, \\forall timestep \\in timesteps

    Where :math:`timestep_{final}` is the last timestep of the timeseries
    """

    storage_initial = get_param(backend_model, "storage_initial", loc_tech)

    storage_loss = get_param(backend_model, "storage_loss", loc_tech)
    if hasattr(backend_model, "storage_inter_cluster"):
        storage = backend_model.storage_inter_cluster
        final_step = backend_model.datesteps[-1]
        time_resolution = 24
    else:
        storage = backend_model.storage
        final_step = backend_model.timesteps[-1]
        time_resolution = backend_model.timestep_resolution[final_step]

    return (
        storage[loc_tech, final_step] * ((1 - storage_loss) ** time_resolution)
        == storage_initial * backend_model.storage_cap[loc_tech]
    )



import pytest  # pylint: disable=unused-import
from pytest import approx
import pandas as pd

import calliope
import calliope.exceptions as exceptions

from calliope.test.common.util import build_test_model as build_model
from calliope.test.common.util import check_error_or_warning


@pytest.fixture(scope="class")
def model():
    m = build_model({}, "simple_supply,two_hours,investment_costs")
    m.run()
    return m


class TestGetInputParams:
    def test_get_input_params(self, model):
        """
        Test that the function access_model_inputs works
        """

        inputs = model.backend.access_model_inputs()
        assert set(model.inputs.data_vars).symmetric_difference(inputs.data_vars) == {
            "objective_cost_class"
        }


class TestUpdateParam:
    def test_update_param_single_dim(self, model):
        """
        test that the function update_param works with a single dimension
        """

        model.backend.update_param("energy_cap_max", {"1::test_supply_elec": 20})

        assert (
            model._backend_model.energy_cap_max.extract_values()["1::test_supply_elec"]
            == 20
        )

    def test_update_param_multiple_vals(self, model):
        """
        test that the function update_param works with a single dimension
        """

        model.backend.update_param(
            "energy_cap_max", {"1::test_supply_elec": 20, "0::test_supply_elec": 30}
        )

        assert (
            model._backend_model.energy_cap_max.extract_values()["1::test_supply_elec"]
            == 20
        )
        assert (
            model._backend_model.energy_cap_max.extract_values()["0::test_supply_elec"]
            == 30
        )

    def test_update_param_multiple_dim(self, model):
        """
        test that the function update_param works with multiple dimensions
        """

        model.backend.update_param(
            "resource",
            {("0::test_demand_elec", "2005-01-01 01:00"): -10},
        )

        assert (
            model._backend_model.resource.extract_values()[
                ("0::test_demand_elec", "2005-01-01 01:00")
            ]
            == -10
        )

    def test_unknown_param(self, model):
        """
        Raise error on unknown param
        """

        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.update_param("unknown_param", {("1::test_supply_elec"): 20})

        assert check_error_or_warning(
            excinfo, "Parameter `unknown_param` not in the Pyomo Backend."
        )

    def test_not_a_param(self, model):
        """
        Raise error when trying to update a non-Param Pyomo object
        """
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.update_param("energy_cap", {"1::test_supply_elec": 20})

        assert check_error_or_warning(
            excinfo, "`energy_cap` not a Parameter in the Pyomo Backend."
        )

        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.update_param("loc_techs", {"1::test_supply_elec": 20})

        assert check_error_or_warning(
            excinfo, "`loc_techs` not a Parameter in the Pyomo Backend."
        )

    def index_not_in_param(self, model):
        """
        Raise error when accessing unknown index
        """

        with pytest.raises(KeyError, match=r"Index 'region1-xc1::csp'"):
            model.backend.update_param("energy_cap_max", {"2::test_supply_elec": 20})

        with pytest.raises(KeyError, match=r"Index 'region1-xc1::csp'"):
            model.backend.update_param(
                "energy_cap_max", {"1::test_supply_elec": 20, "2::test_supply_elec": 20}
            )


class TestActivateConstraint:
    def test_activate_constraint(self, model):
        """
        test that the function activate_constraint works
        """
        model.backend.activate_constraint("system_balance_constraint", active=False)
        assert not model._backend_model.system_balance_constraint.active

        model.backend.activate_constraint("system_balance_constraint", active=True)
        assert model._backend_model.system_balance_constraint.active

    def test_fail_on_activate_unknown_constraint(self, model):
        """
        test that the function activate_constraint fails if unknown constraint
        """
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.activate_constraint("unknown_constraint", active=False)

        assert check_error_or_warning(
            excinfo,
            "constraint/objective `unknown_constraint` not in the Pyomo Backend.",
        )

    def test_fail_on_parameter_activate(self, model):
        """
        test that the function activate_constraint fails if trying to activate a
        non-constraint Pyomo object.
        """
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.activate_constraint("resource", active=False)

        assert check_error_or_warning(
            excinfo, "`resource` not a constraint in the Pyomo Backend."
        )

    def test_non_boolean_parameter_activate(self, model):
        """
        test that the function activate_constraint fails when setting active to
        non-boolean
        """
        with pytest.raises(ValueError) as excinfo:
            model.backend.activate_constraint("system_balance_constraint", active=None)

        assert check_error_or_warning(
            excinfo, "Argument `active` must be True or False"
        )


class TestBackendRerun:
    def test_rerun(self, model):
        """
        test that the function rerun works
        """
        with pytest.warns(exceptions.ModelWarning) as excinfo:
            new_model = model.backend.rerun()

        assert isinstance(new_model, calliope.Model)
        for i in ["_timings", "inputs", "results"]:
            assert hasattr(new_model, i)

        # In new_model.inputs all NaN values have been replace by their default,
        # so we can't directly compare new_model.inputs and model.inputs
        assert new_model.inputs.equals(
            model.backend.access_model_inputs().reindex(new_model.inputs.coords)
        )

        assert new_model.results.equals(model.results)

        assert check_error_or_warning(
            excinfo, "The results of rerunning the backend model are only available"
        )

    def test_update_and_rerun(self, model):
        """
        test that the function rerun works
        """
        model.backend.update_param("energy_cap_max", {"1::test_supply_elec": 20})
        with pytest.warns(exceptions.ModelWarning) as excinfo:
            new_model = model.backend.rerun()

        assert (
            new_model.inputs.energy_cap_max.loc[{"loc_techs": "1::test_supply_elec"}]
            == 20
        )

        assert check_error_or_warning(
            excinfo, "The results of rerunning the backend model are only available"
        )

    def test_rerun_spores(self, model):
        model = calliope.examples.national_scale(
            override_dict={
                "model.subset_time": ["2005-01-01", "2005-01-03"],
                "run.solver": "cbc",
            },
            scenario="spores",
        )

        model.run(build_only=True)
        new_model = model.backend.rerun()
        for i in ["_timings", "inputs", "results"]:
            assert hasattr(new_model, i)
        assert "spores" in new_model.results.dims

    def test_rerun_spores_fail_on_rerun_with_results(self, model):
        model = calliope.examples.national_scale(
            override_dict={
                "model.subset_time": ["2005-01-01", "2005-01-03"],
                "run.solver": "cbc",
            },
            scenario="spores",
        )

        model.run()
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.rerun()
        assert check_error_or_warning(
            excinfo, "Cannot run SPORES if the backend model already has a solution"
        )

    def test_rerun_fail_on_operate(self, model):
        # should fail if the run mode is not 'plan'
        model.run_config["mode"] = "operate"
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.rerun()
        assert check_error_or_warning(
            excinfo, "Cannot rerun the backend in operate run mode"
        )


class TestGetAllModelAttrs:
    def test_get_all_attrs(self, model):
        """Model attributes consist of variables, parameters, and sets"""
        attrs = model.backend.get_all_model_attrs()

        assert attrs.keys() == set(["Set", "Param", "Var"])
        assert isinstance(attrs["Var"], dict)
        assert isinstance(attrs["Param"], dict)
        assert isinstance(attrs["Set"], list)

    def test_check_attrs(self, model):
        """Test one of each object type, just to make sure they are correctly assigned"""
        attrs = model.backend.get_all_model_attrs()

        assert "energy_cap" in attrs["Var"].keys()
        assert "resource" in attrs["Param"].keys()
        assert "carriers" in attrs["Set"]


class TestAddConstraint:
    def test_no_backend(self, model):
        """Must include 'backend_model' as first function argument"""

        def energy_cap_time_varying_rule(backend, loc_tech, timestep):

            return (
                backend.energy_cap[loc_tech]
                <= backend.energy_cap[loc_tech] * backend.resource[loc_tech, timestep]
            )

        constraint_name = "energy_cap_time_varying"
        constraint_sets = ["loc_techs_finite_resource", "timesteps"]
        with pytest.raises(AssertionError) as excinfo:
            model.backend.add_constraint(
                constraint_name, constraint_sets, energy_cap_time_varying_rule
            )
        assert check_error_or_warning(
            excinfo, "First argument of constraint function must be 'backend_model'."
        )

    def test_arg_mismatch(self, model):
        """length of function arguments = length of sets + 1"""

        def energy_cap_time_varying_rule(backend_model, loc_tech, timestep, extra_arg):

            return (
                backend_model.energy_cap[loc_tech]
                <= backend_model.energy_cap[loc_tech]
                * backend_model.resource[loc_tech, timestep]
                + extra_arg
            )

        constraint_name = "energy_cap_time_varying"
        constraint_sets = ["loc_techs_finite_resource", "timesteps"]
        with pytest.raises(AssertionError) as excinfo:
            model.backend.add_constraint(
                constraint_name, constraint_sets, energy_cap_time_varying_rule
            )
        assert check_error_or_warning(
            excinfo,
            "Number of constraint arguments must equal number of constraint sets + 1.",
        )

    def test_sets(self, model):
        """Constraint sets must be backend model sets"""

        def energy_cap_time_varying_rule(backend_model, loc_tech, not_a_set):

            return (
                backend_model.energy_cap[loc_tech]
                <= backend_model.energy_cap[loc_tech]
                * backend_model.resource[loc_tech, not_a_set]
            )

        constraint_name = "energy_cap_time_varying"
        constraint_sets = ["loc_techs_finite_resource", "not_a_set"]
        with pytest.raises(AttributeError) as excinfo:
            model.backend.add_constraint(
                constraint_name, constraint_sets, energy_cap_time_varying_rule
            )

        assert check_error_or_warning(
            excinfo, "Pyomo backend model object has no attribute 'not_a_set'"
        )

    def test_added_constraint(self, model):
        """
        Test the successful addition of a constraint which only allows carrier
        consumption at a maximum rate of half the energy capacity.
        """

        def new_constraint_rule(backend_model, loc_tech_carrier, timestep):
            loc_tech = calliope.backend.pyomo.util.get_loc_tech(loc_tech_carrier)
            carrier_con = backend_model.carrier_con[loc_tech_carrier, timestep]
            timestep_resolution = backend_model.timestep_resolution[timestep]
            return carrier_con * 2 >= (
                -1 * backend_model.energy_cap[loc_tech] * timestep_resolution
            )

        constraint_name = "new_constraint"
        constraint_sets = ["loc_tech_carriers_con", "timesteps"]
        model.backend.add_constraint(
            constraint_name, constraint_sets, new_constraint_rule
        )

        assert hasattr(model._backend_model, "new_constraint")

        new_model = model.backend.rerun()

        assert (
            new_model.results.energy_cap.loc["1::test_demand_elec"]
            == model.results.energy_cap.loc["1::test_demand_elec"] * 2
        )


@pytest.mark.filterwarnings(
    "ignore:(?s).*The results of rerunning the backend model:calliope.exceptions.ModelWarning"
)
@pytest.importorskip("gurobipy")
class TestRegeneratePersistentConstraints:
    @pytest.mark.filterwarnings(
        "ignore:(?s).*Updating the Pyomo parameter:calliope.exceptions.ModelWarning"
    )
    def test_opt_exists(self, model_persistent):
        assert hasattr(model_persistent, "_backend_model_opt")
        assert model_persistent._backend_model_opt.name == "gurobi_persistent"

    @pytest.mark.filterwarnings(
        "ignore:(?s).*Updating the Pyomo parameter:calliope.exceptions.ModelWarning"
    )
    def test_update_param_without_regeneration(self, model_persistent):
        model_persistent.backend.update_param(
            "energy_cap_max", {"1::test_supply_elec": 5}
        )
        model2 = model_persistent.backend.rerun()
        assert model2.results.energy_cap.loc[{"loc_techs": "1::test_supply_elec"}] == 10

    @pytest.mark.filterwarnings(
        "ignore:(?s).*Updating the Pyomo parameter:calliope.exceptions.ModelWarning"
    )
    def test_update_param_with_regeneration_one_dim(self, model_persistent):
        model_persistent.backend.update_param(
            "energy_cap_max", {"1::test_supply_elec": 5, "0::test_supply_elec": 5}
        )
        model_persistent.backend.regenerate_persistent_solver(
            constraints={
                "energy_capacity_constraint": [
                    "1::test_supply_elec",
                    "0::test_supply_elec",
                ]
            }
        )
        model2 = model_persistent.backend.rerun()
        for i in ["1::test_supply_elec", "0::test_supply_elec"]:
            assert model2.results.energy_cap.loc[i] == 5

    @pytest.mark.filterwarnings(
        "ignore:(?s).*Updating the Pyomo parameter:calliope.exceptions.ModelWarning"
    )
    def test_update_param_with_regeneration_two_dims(self, model_persistent):
        model_persistent.backend.update_param(
            "resource", {("1::test_demand_elec", "2005-01-01 01:00"): -4}
        )
        model_persistent.backend.regenerate_persistent_solver(
            constraints={
                "balance_demand_constraint": [
                    ("1::test_demand_elec", "2005-01-01 01:00")
                ]
            }
        )
        model2 = model_persistent.backend.rerun()
        assert (
            model_persistent.results.required_resource.loc[
                ("1::test_demand_elec", "2005-01-01 01:00")
            ]
            == -5
        )
        assert (
            model2.results.required_resource.loc[
                ("1::test_demand_elec", "2005-01-01 01:00")
            ]
            == -4
        )

    @pytest.mark.filterwarnings(
        "ignore:(?s).*Updating the Pyomo parameter:calliope.exceptions.ModelWarning"
    )
    def test_update_obj_without_regeneration(self, model_persistent):
        model_persistent.backend.update_param("objective_cost_class", {"monetary": 0.5})
        model2 = model_persistent.backend.rerun()
        assert model2._model_data.attrs["objective_function_value"] == approx(
            model_persistent._model_data.attrs["objective_function_value"]
        )

    @pytest.mark.filterwarnings(
        "ignore:(?s).*Updating the Pyomo parameter:calliope.exceptions.ModelWarning"
    )
    def test_update_obj_with_regeneration(self, model_persistent):
        model_persistent.backend.update_param("objective_cost_class", {"monetary": 0.5})
        model_persistent.backend.regenerate_persistent_solver(obj=True)
        model2 = model_persistent.backend.rerun()
        assert model2._model_data.attrs["objective_function_value"] == approx(
            0.5 * model_persistent._model_data.attrs["objective_function_value"]
        )

    def test_regeneration_needed_warning(self, model_persistent):
        with pytest.warns(exceptions.ModelWarning) as excinfo:
            model_persistent.backend.update_param(
                "objective_cost_class", {"monetary": 0.5}
            )
        assert check_error_or_warning(
            excinfo, "Updating the Pyomo parameter won't affect the optimisation"
        )

    def test_fail_to_regenerate_non_persistent_solver(self, model):
        with pytest.raises(exceptions.ModelError) as excinfo:
            model.backend.regenerate_persistent_solver(obj=True)
        assert check_error_or_warning(excinfo, "Can only regenerate persistent solvers")



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

model.py
~~~~~~~~

Implements the core Model class.

"""

from io import StringIO
import logging
import warnings

import numpy as np
import xarray as xr
import ruamel.yaml as ruamel_yaml

from calliope.postprocess import results as postprocess_results
from calliope.postprocess import plotting
from calliope.core import io
from calliope.preprocess import (
    model_run_from_yaml,
    model_run_from_dict,
    build_model_data,
    apply_time_clustering,
    final_timedimension_processing,
)
from calliope.core.attrdict import AttrDict
from calliope.core.util.logging import log_time
from calliope.core.util.dataset import split_loc_techs
from calliope.core.util.tools import apply_to_dict
from calliope.core.util.observed_dict import UpdateObserverDict
from calliope import exceptions
from calliope.backend.run import run as run_backend

logger = logging.getLogger(__name__)


def read_netcdf(path):
    """
    Return a Model object reconstructed from model data in a NetCDF file.

    """
    model_data = io.read_netcdf(path)
    return Model(config=None, model_data=model_data)


class Model(object):
    """
    A Calliope Model.

    """

    def __init__(self, config, model_data=None, *args, **kwargs):
        """
        Returns a new Model from either the path to a YAML model
        configuration file or a dict fully specifying the model.

        Parameters
        ----------
        config : str or dict or AttrDict
            If str, must be the path to a model configuration file.
            If dict or AttrDict, must fully specify the model.
        model_data : Dataset, optional
            Create a Model instance from a fully built model_data Dataset.
            This is only used if `config` is explicitly set to None
            and is primarily used to re-create a Model instance from
            a model previously saved to a NetCDF file.

        """
        self._timings = {}
        # try to set logging output format assuming python interactive. Will
        # use CLI logging format if model called from CLI
        log_time(logger, self._timings, "model_creation", comment="Model: initialising")
        if isinstance(config, str):
            model_run, debug_data = model_run_from_yaml(config, *args, **kwargs)
            self._init_from_model_run(model_run, debug_data)
        elif isinstance(config, dict):
            model_run, debug_data = model_run_from_dict(config, *args, **kwargs)
            self._init_from_model_run(model_run, debug_data)
        elif model_data is not None and config is None:
            self._init_from_model_data(model_data)
        else:
            # expected input is a string pointing to a YAML file of the run
            # configuration or a dict/AttrDict in which the run and model
            # configurations are defined
            raise ValueError(
                "Input configuration must either be a string or a dictionary."
            )
        self._check_future_deprecation_warnings()

        self.plot = plotting.ModelPlotMethods(self)

    def _init_from_model_run(self, model_run, debug_data):
        self._model_run = model_run
        self._debug_data = debug_data
        log_time(
            logger,
            self._timings,
            "model_run_creation",
            comment="Model: preprocessing stage 1 (model_run)",
        )

        self._model_data_original = build_model_data(model_run)
        log_time(
            logger,
            self._timings,
            "model_data_original_creation",
            comment="Model: preprocessing stage 2 (model_data)",
        )

        random_seed = self._model_run.get_key("model.random_seed", None)
        if random_seed:
            np.random.seed(seed=random_seed)

        # After setting the random seed, time clustering can take place
        time_config = model_run.model.get("time", None)
        if not time_config:
            _model_data = self._model_data_original
        else:
            _model_data = apply_time_clustering(self._model_data_original, model_run)
        self._model_data = final_timedimension_processing(_model_data)
        log_time(
            logger,
            self._timings,
            "model_data_creation",
            comment="Model: preprocessing complete",
        )

        # Ensure model and run attributes of _model_data update themselves
        for var in self._model_data.data_vars:
            self._model_data[var].attrs["is_result"] = 0
        self.inputs = self._model_data.filter_by_attrs(is_result=0)

        model_config = {
            k: v for k, v in model_run.get("model", {}).items() if k != "file_allowed"
        }
        self.model_config = UpdateObserverDict(
            initial_dict=model_config, name="model_config", observer=self._model_data
        )
        self.run_config = UpdateObserverDict(
            initial_dict=model_run.get("run", {}),
            name="run_config",
            observer=self._model_data,
        )

    def _init_from_model_data(self, model_data):
        if "_model_run" in model_data.attrs:
            self._model_run = AttrDict.from_yaml_string(model_data.attrs["_model_run"])
            del model_data.attrs["_model_run"]

        if "_debug_data" in model_data.attrs:
            self._debug_data = AttrDict.from_yaml_string(
                model_data.attrs["_debug_data"]
            )
            del model_data.attrs["_debug_data"]

        self._model_data = model_data
        self._add_model_data_methods()

        log_time(
            logger,
            self._timings,
            "model_data_loaded",
            comment="Model: loaded model_data",
        )

    def _add_model_data_methods(self):
        self.inputs = self._model_data.filter_by_attrs(is_result=0)
        self.results = self._model_data.filter_by_attrs(is_result=1)
        self.model_config = UpdateObserverDict(
            initial_yaml_string=self._model_data.attrs.get("model_config", "{}"),
            name="model_config",
            observer=self._model_data,
        )
        self.run_config = UpdateObserverDict(
            initial_yaml_string=self._model_data.attrs.get("run_config", "{}"),
            name="run_config",
            observer=self._model_data,
        )

    def save_commented_model_yaml(self, path):
        """
        Save a fully built and commented version of the model to a YAML file
        at the given ``path``. Comments in the file indicate where values
        were overridden. This is Calliope's internal representation of
        a model directly before the model_data xarray.Dataset is built,
        and can be useful for debugging possible issues in the model
        formulation.

        """
        if not self._model_run or not self._debug_data:
            raise KeyError(
                "This model does not have the fully built model attached, "
                "so `save_commented_model_yaml` is not available. Likely "
                "reason is that the model was built with a verion of Calliope "
                "prior to 0.6.5."
            )

        yaml = ruamel_yaml.YAML()

        model_run_debug = self._model_run.copy()
        try:
            del model_run_debug["timeseries_data"]  # Can't be serialised!
        except KeyError:
            # Possible that timeseries_data is already gone if the model
            # was read from a NetCDF file
            pass

        # Turn sets in model_run into lists for YAML serialization
        for k, v in model_run_debug.sets.items():
            model_run_debug.sets[k] = list(v)

        debug_comments = self._debug_data["comments"]

        stream = StringIO()
        yaml.dump(model_run_debug.as_dict(), stream=stream)
        debug_yaml = yaml.load(stream.getvalue())

        for k in debug_comments.model_run.keys_nested():
            v = debug_comments.model_run.get_key(k)
            if v:
                keys = k.split(".")
                apply_to_dict(
                    debug_yaml, keys[:-1], "yaml_add_eol_comment", (v, keys[-1])
                )

        dumper = ruamel_yaml.dumper.RoundTripDumper
        dumper.ignore_aliases = lambda self, data: True

        with open(path, "w") as f:
            ruamel_yaml.dump(
                debug_yaml, stream=f, Dumper=dumper, default_flow_style=False
            )

    def run(self, force_rerun=False, **kwargs):
        """
        Run the model. If ``force_rerun`` is True, any existing results
        will be overwritten.

        Additional kwargs are passed to the backend.

        """
        # Check that results exist and are non-empty
        if hasattr(self, "results") and self.results.data_vars and not force_rerun:
            raise exceptions.ModelError(
                "This model object already has results. "
                "Use model.run(force_rerun=True) to force"
                "the results to be overwritten with a new run."
            )

        if (
            self.run_config["mode"] == "operate"
            and not self._model_data.attrs["allow_operate_mode"]
        ):
            raise exceptions.ModelError(
                "Unable to run this model in operational mode, probably because "
                "there exist non-uniform timesteps (e.g. from time masking)"
            )

        results, self._backend_model, self._backend_model_opt, interface = run_backend(
            self._model_data, self._timings, **kwargs
        )

        # Add additional post-processed result variables to results
        if results.attrs.get("termination_condition", None) in ["optimal", "feasible"]:
            results = postprocess_results.postprocess_model_results(
                results, self._model_data, self._timings
            )
        self._model_data.attrs.update(results.attrs)
        self._model_data = xr.merge(
            [results, self._model_data], compat="override", combine_attrs="no_conflicts"
        )
        self._add_model_data_methods()

        self.backend = interface(self)

    def get_formatted_array(self, var, index_format="index"):
        """
        Return an xr.DataArray with locs, techs, and carriers as
        separate dimensions.

        Parameters
        ----------
        var : str
            Decision variable for which to return a DataArray.
        index_format : str, default = 'index'
            'index' to return the `loc_tech(_carrier)` dimensions as individual
            indexes, 'multiindex' to return them as a MultiIndex. The latter
            has the benefit of having a smaller memory footprint, but you cannot
            undertake dimension specific operations (e.g. formatted_array.sum('locs'))
        """

        if var not in self._model_data.data_vars:
            raise KeyError("Variable {} not in Model data".format(var))

        if index_format not in ["index", "multiindex"]:
            raise ValueError(
                "Argument 'index_format' must be one of 'index' or 'multiindex'"
            )
        elif index_format == "index":
            return_as = "DataArray"
        elif index_format == "multiindex":
            return_as = "MultiIndex DataArray"

        return split_loc_techs(self._model_data[var], return_as=return_as)

    def to_netcdf(self, path):
        """
        Save complete model data (inputs and, if available, results)
        to a NetCDF file at the given ``path``.

        """
        io.save_netcdf(self._model_data, path, model=self)

    def to_csv(self, path, dropna=True):
        """
        Save complete model data (inputs and, if available, results)
        as a set of CSV files to the given ``path``.

        Parameters
        ----------
        dropna : bool, optional
            If True (default), NaN values are dropped when saving,
            resulting in significantly smaller CSV files.

        """
        io.save_csv(self._model_data, path, dropna)

    def to_lp(self, path):
        """
        Save built model to LP format at the given ``path``. If the backend
        model has not been built yet, it is built prior to saving.
        """
        io.save_lp(self, path)

    def info(self):
        info_strings = []
        model_name = self.model_config.get("name", "None")
        info_strings.append("Model name:   {}".format(model_name))
        msize = "{locs} locations, {techs} technologies, {times} timesteps".format(
            locs=len(self._model_data.coords.get("locs", [])),
            techs=(
                len(self._model_data.coords.get("techs_non_transmission", []))
                + len(self._model_data.coords.get("techs_transmission_names", []))
            ),
            times=len(self._model_data.coords.get("timesteps", [])),
        )
        info_strings.append("Model size:   {}".format(msize))
        return "\n".join(info_strings)

    def _check_future_deprecation_warnings(self):
        """
        Method for all FutureWarnings and DeprecationWarnings. Comment above each
        warning should specify Calliope version in which it was added, and the
        version in which it should be updated/removed.
        """

        # Warning that group_share constraints will removed in 0.7.0 #
        # Added in 0.6.4-dev, to be removed in v0.7.0-dev
        if any("group_share_" in i for i in self._model_data.data_vars.keys()):
            warnings.warn(
                "`group_share` constraints will be removed in v0.7.0 -- "
                "use the new model-wide constraints instead.",
                FutureWarning,
            )

        # Warning that charge rate will be removed in 0.7.0
        # Added in 0.6.4-dev, to be removed in 0.7.0-dev
        # Rename charge rate to energy_cap_per_storage_cap_max
        if self._model_data is not None and "charge_rate" in self._model_data:
            warnings.warn(
                "`charge_rate` is renamed to `energy_cap_per_storage_cap_max` "
                "and will be removed in v0.7.0.",
                FutureWarning,
            )



import os
import sys

import pytest
from pyomo.core.expr.current import identify_variables

import calliope
from calliope import AttrDict


constraint_sets = AttrDict.from_yaml(
    os.path.join(os.path.dirname(__file__), "constraint_sets.yaml")
)

defaults = AttrDict.from_yaml(
    os.path.join(os.path.dirname(calliope.__file__), "config", "defaults.yaml")
)


python36_or_higher = pytest.mark.skipif(
    sys.version_info < (3, 6), reason="Requires ordered dicts from Python >= 3.6"
)


def build_test_model(
    override_dict=None,
    scenario=None,
    model_file="model.yaml",
    timeseries_dataframes=None,
):
    return calliope.Model(
        os.path.join(os.path.dirname(__file__), "test_model", model_file),
        override_dict=override_dict,
        scenario=scenario,
        timeseries_dataframes=timeseries_dataframes,
    )


def check_error_or_warning(error_warning, test_string_or_strings):
    if hasattr(error_warning, "list"):
        output = ",".join(
            str(error_warning.list[i]) for i in range(len(error_warning.list))
        )
    else:
        output = str(error_warning.value)

    if isinstance(test_string_or_strings, list):
        result = all(test_string in output for test_string in test_string_or_strings)
    else:
        result = test_string_or_strings in output

    return result


def check_variable_exists(backend_model, constraint, variable):
    """
    Search for existence of a decision variable in a Pyomo constraint.

    Parameters
    ----------
    backend_model : Pyomo ConcreteModel
    constraint : str, name of constraint which could exist in the backend
    variable : str, string to search in the list of variables to check if existing
    """
    exists = []
    for v in getattr(backend_model, constraint).values():
        variables = identify_variables(v.body)
        exists.append(any(variable in j.getname() for j in list(variables)))
    return any(exists)


def get_indexed_constraint_body(backend_model, constraint, input_index):
    """
    Return all indeces of a specific decision variable used in a constraint.
    This is useful to check that all expected loc_techs are in a summation.

    Parameters
    ----------
    backend_model : Pyomo ConcreteModel
    constraint : str,
        Name of constraint which could exist in the backend
    input_index : tuple or string,
        The index of the constraint in which to look for a the variable.
        The index may impact the index of the decision ariable

    """
    constraint_index = [
        v
        for v in getattr(backend_model, constraint).values()
        if v.index() == input_index
    ]
    if len(constraint_index) == 0:
        raise KeyError(
            "Unable to find index {} in constraint {}".format(input_index, constraint)
        )
    else:
        return constraint_index[0].body



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

funcs.py
~~~~~~~~

Functions to process time series data.

"""

import logging

import numpy as np
import pandas as pd
import xarray as xr

from calliope import exceptions
from calliope.core.util.dataset import get_loc_techs
from calliope.time import clustering
from calliope.preprocess.lookup import lookup_clusters

logger = logging.getLogger(__name__)


def get_daily_timesteps(data, check_uniformity=False):
    daily_timesteps = [
        data.timestep_resolution.loc[i].values
        for i in np.unique(data.timesteps.to_index().strftime("%Y-%m-%d"))
    ]

    if check_uniformity:
        if not np.all(daily_timesteps == daily_timesteps[0]):
            raise exceptions.ModelError(
                "For clustering, timestep resolution must be uniform."
            )

    return daily_timesteps[0]


def normalized_copy(data):
    """
    Normalize timeseries data, using the maximum across all regions and timesteps.

    Parameters
    ----------
    data : xarray Dataset
        Dataset with all non-time dependent variables removed

    Returns
    -------
    ds : xarray Dataset
        Copy of `data`, with the absolute taken and normalized to 0-1

    """
    ds = data.copy(deep=True)  # Work off a copy

    for var in ds.data_vars:
        # Each DataArray is indexed over a different subset of loc_techs,
        # so we find it in the list of dimensions
        loc_tech_dim = [i for i in ds[var].dims if "loc_techs" in i][0]

        # For each technology, get the loc_techs which are relevant
        loc_tech_subsets = [
            get_loc_techs(ds[loc_tech_dim].values, tech)
            for tech in set(i.split("::")[1] for i in ds[loc_tech_dim].values)
        ]
        # remove empty lists within the _techs list
        loc_tech_subsets = [i for i in loc_tech_subsets if i]

        # For each technology, divide all values by the maximum absolute value
        for loc_tech in loc_tech_subsets:
            ds[var].loc[{loc_tech_dim: loc_tech}] = abs(
                ds[var].loc[{loc_tech_dim: loc_tech}]
                / abs(ds[var].loc[{loc_tech_dim: loc_tech}]).max()
            )
    return ds


def _copy_non_t_vars(data0, data1):
    """Copies non-t-indexed variables from data0 into data1, then
    returns data1"""
    non_t_vars = [
        varname
        for varname, vardata in data0.data_vars.items()
        if "timesteps" not in vardata.dims
    ]
    # Manually copy over variables not in `timesteps`. If we don't do this,
    # these vars get polluted with a superfluous `timesteps` dimension
    for v in non_t_vars:
        data1[v] = data0[v]
    return data1


def _combine_datasets(data0, data1):
    """Concatenates data0 and data1 along the time dimension"""
    data_new = xr.concat([data0, data1], dim="timesteps")
    # Ensure time dimension is ordered
    data_new = data_new.loc[{"timesteps": data_new.timesteps.to_index().sort_values()}]

    return data_new


def _drop_timestep_vars(data, timesteps):
    timeseries_data = data.copy(deep=True)
    # Save all coordinates, to ensure they can be added back in after clustering
    data_coords = data.copy().coords
    del data_coords["timesteps"]

    if timesteps is not None:
        timeseries_data = timeseries_data.loc[{"timesteps": timesteps}]

    timeseries_data = timeseries_data.drop_vars(
        [
            varname
            for varname, vardata in data.data_vars.items()
            if "timesteps" not in vardata.dims
        ]
    )

    return timeseries_data, data_coords


def apply_clustering(
    data,
    timesteps,
    clustering_func,
    how,
    normalize=True,
    scale_clusters="mean",
    storage_inter_cluster=True,
    model_run=None,
    **kwargs,
):
    """
    Apply the given clustering function to the given data.

    Parameters
    ----------
    data : xarray.Dataset
    timesteps : pandas.DatetimeIndex or list of timesteps or None
    clustering_func : str
        Name of clustering function. Can be `file=....csv:column_name`
        if loading custom clustering. Custom clustering index = timeseries days.
        If no column_name, the CSV file must have only one column of data.
    how : str
        How to map clusters to data. 'mean' or 'closest'.
    normalize : bool, optional
        If True (default), data is normalized before clustering is applied,
        using :func:`~calliope.time.funcs.normalized_copy`.
    scale_clusters : str or None, default = 'mean'
        Scale the results of clustering such that the clusters match the metric
        given by scale_clusters. For example, 'mean' scales along each loc_tech
        and variable to match inputs and outputs. Other options for matching
        include 'sum', 'max', and 'min'. If None, no scaling occurs.
    **kwargs : optional
        Arguments passed to clustering_func.

    Returns
    -------
    data_new_scaled : xarray.Dataset

    """

    assert how in ["mean", "closest"]

    daily_timesteps = get_daily_timesteps(data, check_uniformity=True)
    timesteps_per_day = len(daily_timesteps)

    # get a copy of the dataset with only timeseries variables,
    # and get all coordinates of the original dataset, to reinstate later
    data_to_cluster, data_coords = _drop_timestep_vars(data, timesteps)

    data_to_cluster = data_to_cluster.drop_vars(
        ["timestep_weights", "timestep_resolution"]
    )

    for dim in data_to_cluster.dims:
        data_to_cluster[dim] = data[dim]

    if normalize:
        data_normalized = normalized_copy(data_to_cluster)
    else:
        data_normalized = data_to_cluster

    if "file=" in clustering_func:
        file = clustering_func.split("=")[1]
        if ":" in file:
            file, column = file.rsplit(":", 1)
        else:
            column = None

        df = model_run.timeseries_data[file]
        if isinstance(df, pd.Series) and column is not None:
            exceptions.warn(
                "{} given as time clustering column, but only one column to "
                "choose from in {}.".format(column, file)
            )
            clusters = df.resample("1D").mean()
        elif isinstance(df, pd.DataFrame) and column is None:
            raise exceptions.ModelError(
                "No time clustering column given, but multiple columns found in "
                "{0}. Choose one column and add it to {1} as {1}:name_of_column.".format(
                    file, clustering_func
                )
            )
        elif isinstance(df, pd.DataFrame) and column not in df.columns:
            raise KeyError(
                "time clustering column {} not found in {}.".format(column, file)
            )
        elif isinstance(df, pd.DataFrame):
            clusters = df.loc[:, column].groupby(pd.Grouper(freq="1D")).unique()

        # Check there weren't instances of more than one cluster assigned to a day
        # or days with no information assigned
        if any([len(i) == 0 for i in clusters.values]):
            raise exceptions.ModelError(
                "Missing cluster days in `{}:{}`.".format(file, column)
            )
        elif any([len(i) > 1 for i in clusters.values]):
            raise exceptions.ModelError(
                "More than one cluster value assigned to a day in `{}:{}`. "
                "Unique clusters per day: {}".format(file, column, clusters)
            )
        else:
            clusters.loc[:] = [i[0] for i in clusters.values]

    else:
        result = clustering.get_clusters(
            data_normalized,
            clustering_func,
            timesteps_per_day=timesteps_per_day,
            **kwargs,
        )
        clusters = result[0]  # Ignore other stuff returned

    data_new = clustering.map_clusters_to_data(
        data_to_cluster,
        clusters,
        how=how,
        daily_timesteps=daily_timesteps,
        storage_inter_cluster=storage_inter_cluster,
    )

    # It's now safe to add the original coordinates back in (preserving all the
    # loc_tech sets that aren't used to index a variable in the DataArray)
    data_new.update(data_coords)

    data_new = _copy_non_t_vars(data, data_new)

    if timesteps is not None:
        data_new = _copy_non_t_vars(data, data_new)
        data_new = _combine_datasets(data.drop_sel(timesteps=timesteps), data_new)
        data_new = _copy_non_t_vars(data, data_new)

    # Scale the new/combined data so that the mean for each (loc_tech, variable)
    # combination matches that from the original data
    data_new_scaled = data_new.copy(deep=True)
    if scale_clusters:
        data_vars_in_t = [
            v
            for v in data_new.data_vars
            if "timesteps" in data_new[v].dims
            and "timestep_" not in v
            and v != "clusters"
        ]
        for var in data_vars_in_t:
            scale = getattr(data[var], scale_clusters)(dim="timesteps") / getattr(
                data_new[var], scale_clusters
            )(dim="timesteps")
            data_new_scaled[var] = data_new[var] * scale.fillna(0)

    lookup_clusters(data_new_scaled)

    return data_new_scaled


def resample(data, timesteps, resolution):
    """
    Function to resample timeseries data from the input resolution (e.g. 1H), to
    the given resolution (e.g. 2H)

    Parameters
    ----------
    data : xarray.Dataset
        calliope model data, containing only timeseries data variables
    timesteps : str or list; optional
        If given, apply resampling to a subset of the timeseries data
    resolution : str
        time resolution of the output data, given in Pandas time frequency format.
        E.g. 1H = 1 hour, 1W = 1 week, 1M = 1 month, 1T = 1 minute. Multiples allowed.

    """

    def _resample(var, how):
        return getattr(var.resample(timesteps=resolution, keep_attrs=True), how)(
            "timesteps"
        )

    # get a copy of the dataset with only timeseries variables,
    # and get all coordinates of the original dataset, to reinstate later
    data_new, data_coords = _drop_timestep_vars(data, timesteps)

    # First create a new resampled dataset of the correct size by
    # using first-resample, which should be a quick way to achieve this

    data_rs = _resample(data_new, how="first")

    for var in data_rs.data_vars:
        if var in ["timestep_resolution", "resource"]:
            data_rs[var] = _resample(data_new[var], how="sum")
        else:
            try:
                data_rs[var] = _resample(data_new[var], how="mean")
            except TypeError:
                # If the var has a datatype of strings, it can't be resampled
                logger.error(
                    "Dropping {} because it has a {} data type when integer or "
                    "float is expected for timeseries resampling.".format(
                        var, data_rs[var].dtype
                    )
                )
                data_rs = data_rs.drop_vars(var)

    # Get rid of the filled-in NaN timestamps
    data_rs = data_rs.dropna(dim="timesteps", how="all")

    data_rs.attrs["allow_operate_mode"] = 1  # Resampling still permits operational mode

    # It's now safe to add the original coordinates back in (preserving all the
    # loc_tech sets that aren't used to index a variable in the DataArray)
    data_rs.update(data_coords)

    data_rs = _copy_non_t_vars(data, data_rs)  # add back in non timeseries data

    if timesteps is not None:
        # Combine leftover parts of passed in data with new data
        data_rs = _combine_datasets(data.drop_sel(timesteps=timesteps), data_rs)
        data_rs = _copy_non_t_vars(data, data_rs)
        # Having timesteps with different lengths does not permit operational mode
        data_rs.attrs["allow_operate_mode"] = 0

    return data_rs


def drop(data, timesteps):
    """
    Drop timesteps from data, adjusting the timestep weight of remaining
    timesteps accordingly. Returns updated dataset.

    Parameters
    ----------
    data : xarray.Dataset
        Calliope model data.
    timesteps : str or list or other iterable
        Pandas-compatible timestep strings.

    """
    # Turn timesteps into a pandas datetime index for subsetting, which also
    # checks whether they are actually valid
    try:
        timesteps_pd = pd.to_datetime(timesteps)
    except Exception as e:
        raise exceptions.ModelError("Invalid timesteps: {}".format(timesteps))

    # 'Distribute weight' of the dropped timesteps onto the remaining ones
    dropped_weight = data.timestep_weights.loc[{"timesteps": timesteps_pd}].sum()

    data = data.drop_sel(timesteps=timesteps_pd)

    data["timestep_weights"] = data["timestep_weights"] + (
        dropped_weight / len(data["timestep_weights"])
    )

    return data



from calliope.postprocess.plotting.plotting import plot_summary, ModelPlotMethods
from calliope.postprocess.plotting.capacity import plot_capacity
from calliope.postprocess.plotting.timeseries import plot_timeseries
from calliope.postprocess.plotting.transmission import plot_transmission
from calliope.postprocess.plotting.flows import plot_flows



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

dispatch.py
~~~~~~~~~~~~~~~~~

Energy dispatch constraints, limiting production/consumption to the capacities
of the technologies

"""

import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import get_param, get_loc_tech, get_previous_timestep

ORDER = 10  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]

    if "loc_tech_carriers_carrier_production_max_constraint" in sets:
        backend_model.carrier_production_max_constraint = po.Constraint(
            backend_model.loc_tech_carriers_carrier_production_max_constraint,
            backend_model.timesteps,
            rule=carrier_production_max_constraint_rule,
        )
    if "loc_tech_carriers_carrier_production_min_constraint" in sets:
        backend_model.carrier_production_min_constraint = po.Constraint(
            backend_model.loc_tech_carriers_carrier_production_min_constraint,
            backend_model.timesteps,
            rule=carrier_production_min_constraint_rule,
        )
    if "loc_tech_carriers_carrier_consumption_max_constraint" in sets:
        backend_model.carrier_consumption_max_constraint = po.Constraint(
            backend_model.loc_tech_carriers_carrier_consumption_max_constraint,
            backend_model.timesteps,
            rule=carrier_consumption_max_constraint_rule,
        )

    if "loc_techs_resource_max_constraint" in sets:
        backend_model.resource_max_constraint = po.Constraint(
            backend_model.loc_techs_resource_max_constraint,
            backend_model.timesteps,
            rule=resource_max_constraint_rule,
        )

    if "loc_techs_storage_max_constraint" in sets:
        backend_model.storage_max_constraint = po.Constraint(
            backend_model.loc_techs_storage_max_constraint,
            backend_model.timesteps,
            rule=storage_max_constraint_rule,
        )

    if "loc_techs_storage_discharge_depth" in sets:
        backend_model.storage_discharge_depth_constraint = po.Constraint(
            backend_model.loc_techs_storage_discharge_depth,
            backend_model.timesteps,
            rule=storage_discharge_depth_constraint_rule,
        )

    if "loc_tech_carriers_ramping_constraint" in sets:
        backend_model.ramping_up_constraint = po.Constraint(
            backend_model.loc_tech_carriers_ramping_constraint,
            backend_model.timesteps,
            rule=ramping_up_constraint_rule,
        )

        backend_model.ramping_down_constraint = po.Constraint(
            backend_model.loc_tech_carriers_ramping_constraint,
            backend_model.timesteps,
            rule=ramping_down_constraint_rule,
        )

    if "loc_techs_storage_intra_max_constraint" in sets:
        backend_model.storage_intra_max_constraint = po.Constraint(
            backend_model.loc_techs_storage_intra_max_constraint,
            backend_model.timesteps,
            rule=storage_intra_max_rule,
        )
    if "loc_techs_storage_intra_min_constraint" in sets:
        backend_model.storage_intra_min_constraint = po.Constraint(
            backend_model.loc_techs_storage_intra_min_constraint,
            backend_model.timesteps,
            rule=storage_intra_min_rule,
        )
    if "loc_techs_storage_inter_max_constraint" in sets:
        backend_model.storage_inter_max_constraint = po.Constraint(
            backend_model.loc_techs_storage_inter_max_constraint,
            backend_model.datesteps,
            rule=storage_inter_max_rule,
        )
    if "loc_techs_storage_inter_min_constraint" in sets:
        backend_model.storage_inter_min_constraint = po.Constraint(
            backend_model.loc_techs_storage_inter_min_constraint,
            backend_model.datesteps,
            rule=storage_inter_min_rule,
        )


def carrier_production_max_constraint_rule(backend_model, loc_tech_carrier, timestep):
    """
    Set maximum carrier production. All technologies.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep) \\leq energy_{cap}(loc::tech)
            \\times timestep\\_resolution(timestep) \\times parasitic\\_eff(loc::tec)

    """
    loc_tech = get_loc_tech(loc_tech_carrier)
    carrier_prod = backend_model.carrier_prod[loc_tech_carrier, timestep]
    timestep_resolution = backend_model.timestep_resolution[timestep]
    parasitic_eff = get_param(backend_model, "parasitic_eff", (loc_tech, timestep))

    return carrier_prod <= (
        backend_model.energy_cap[loc_tech] * timestep_resolution * parasitic_eff
    )


def carrier_production_min_constraint_rule(backend_model, loc_tech_carrier, timestep):
    """
    Set minimum carrier production. All technologies except ``conversion_plus``.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep) \\geq energy_{cap}(loc::tech)
            \\times timestep\\_resolution(timestep) \\times energy_{cap,min\\_use}(loc::tec)

    """
    loc_tech = get_loc_tech(loc_tech_carrier)
    carrier_prod = backend_model.carrier_prod[loc_tech_carrier, timestep]
    timestep_resolution = backend_model.timestep_resolution[timestep]
    min_use = get_param(backend_model, "energy_cap_min_use", (loc_tech, timestep))

    return carrier_prod >= (
        backend_model.energy_cap[loc_tech] * timestep_resolution * min_use
    )


def carrier_consumption_max_constraint_rule(backend_model, loc_tech_carrier, timestep):
    """
    Set maximum carrier consumption for demand, storage, and transmission techs.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep) \\geq
            -1 \\times energy_{cap}(loc::tech)
            \\times timestep\\_resolution(timestep)

    """
    loc_tech = get_loc_tech(loc_tech_carrier)
    carrier_con = backend_model.carrier_con[loc_tech_carrier, timestep]
    timestep_resolution = backend_model.timestep_resolution[timestep]

    return carrier_con >= (
        -1 * backend_model.energy_cap[loc_tech] * timestep_resolution
    )


def resource_max_constraint_rule(backend_model, loc_tech, timestep):
    """
    Set maximum resource consumed by supply_plus techs.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{con}}(loc::tech, timestep) \\leq
            timestep\\_resolution(timestep) \\times resource_{cap}(loc::tech)

    """
    timestep_resolution = backend_model.timestep_resolution[timestep]

    return backend_model.resource_con[loc_tech, timestep] <= (
        timestep_resolution * backend_model.resource_cap[loc_tech]
    )


def storage_max_constraint_rule(backend_model, loc_tech, timestep):
    """
    Set maximum stored energy. Supply_plus & storage techs only.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep) \\leq
            storage_{cap}(loc::tech)

    """
    return (
        backend_model.storage[loc_tech, timestep] <= backend_model.storage_cap[loc_tech]
    )


def storage_discharge_depth_constraint_rule(backend_model, loc_tech, timestep):
    """
    Forces storage state of charge to be greater than the allowed depth of discharge.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep) >=
            \\boldsymbol{storage_discharge_depth}\\forall loc::tech \\in loc::techs_{storage}, \\forall timestep \\in timesteps

    """
    storage_discharge_depth = get_param(
        backend_model, "storage_discharge_depth", loc_tech
    )
    return (
        backend_model.storage[loc_tech, timestep]
        >= storage_discharge_depth * backend_model.storage_cap[loc_tech]
    )


def ramping_up_constraint_rule(backend_model, loc_tech_carrier, timestep):
    """
    Ramping up constraint.

    .. container:: scrolling-wrapper

        .. math::

            diff(loc::tech::carrier, timestep) \\leq max\\_ramping\\_rate(loc::tech::carrier, timestep)

    """
    return ramping_constraint(backend_model, loc_tech_carrier, timestep, direction=0)


def ramping_down_constraint_rule(backend_model, loc_tech_carrier, timestep):
    """
    Ramping down constraint.

    .. container:: scrolling-wrapper

        .. math::

            -1 \\times max\\_ramping\\_rate(loc::tech::carrier, timestep) \\leq diff(loc::tech::carrier, timestep)

    """
    return ramping_constraint(backend_model, loc_tech_carrier, timestep, direction=1)


def ramping_constraint(backend_model, loc_tech_carrier, timestep, direction=0):
    """
    Ramping rate constraints.

    Direction: 0 is up, 1 is down.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{max\\_ramping\\_rate}(loc::tech::carrier, timestep) =
            energy_{ramping}(loc::tech, timestep) \\times energy_{cap}(loc::tech)

            \\boldsymbol{diff}(loc::tech::carrier, timestep) =
            (carrier_{prod}(loc::tech::carrier, timestep) + carrier_{con}(loc::tech::carrier, timestep))
            / timestep\\_resolution(timestep) -
            (carrier_{prod}(loc::tech::carrier, timestep-1) + carrier_{con}(loc::tech::carrier, timestep-1))
            / timestep\\_resolution(timestep-1)

    """

    # No constraint for first timestep
    # Pyomo returns the order 1-indexed, but we want 0-indexing
    if backend_model.timesteps.ord(timestep) - 1 == 0:
        return po.Constraint.NoConstraint
    else:
        previous_step = get_previous_timestep(backend_model.timesteps, timestep)
        time_res = backend_model.timestep_resolution[timestep]
        time_res_prev = backend_model.timestep_resolution[previous_step]
        loc_tech = loc_tech_carrier.rsplit("::", 1)[0]
        # Ramping rate (fraction of installed capacity per hour)
        ramping_rate = get_param(backend_model, "energy_ramping", (loc_tech, timestep))

        try:
            prod_this = backend_model.carrier_prod[loc_tech_carrier, timestep]
            prod_prev = backend_model.carrier_prod[loc_tech_carrier, previous_step]
        except KeyError:
            prod_this = 0
            prod_prev = 0

        try:
            con_this = backend_model.carrier_con[loc_tech_carrier, timestep]
            con_prev = backend_model.carrier_con[loc_tech_carrier, previous_step]
        except KeyError:
            con_this = 0
            con_prev = 0

        diff = (prod_this + con_this) / time_res - (
            prod_prev + con_prev
        ) / time_res_prev

        max_ramping_rate = ramping_rate * backend_model.energy_cap[loc_tech]

        if direction == 0:
            return diff <= max_ramping_rate
        else:
            return -1 * max_ramping_rate <= diff


def storage_intra_max_rule(backend_model, loc_tech, timestep):
    """
    When clustering days, to reduce the timeseries length, set limits on
    intra-cluster auxiliary maximum storage decision variable.
    `Ref: DOI 10.1016/j.apenergy.2018.01.023 <https://doi.org/10.1016/j.apenergy.2018.01.023>`_

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep) \\leq
            \\boldsymbol{storage_{intra\\_cluster, max}}(loc::tech, cluster(timestep))
            \\quad \\forall loc::tech \\in loc::techs_{store}, \\forall timestep \\in timesteps

    Where :math:`cluster(timestep)` is the cluster number in which the timestep
    is located.
    """
    cluster = backend_model.__calliope_model_data["data"]["timestep_cluster"][timestep]
    return (
        backend_model.storage[loc_tech, timestep]
        <= backend_model.storage_intra_cluster_max[loc_tech, cluster]
    )


def storage_intra_min_rule(backend_model, loc_tech, timestep):
    """
    When clustering days, to reduce the timeseries length, set limits on
    intra-cluster auxiliary minimum storage decision variable.
    `Ref: DOI 10.1016/j.apenergy.2018.01.023 <https://doi.org/10.1016/j.apenergy.2018.01.023>`_

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage}(loc::tech, timestep) \\geq
            \\boldsymbol{storage_{intra\\_cluster, min}}(loc::tech, cluster(timestep))
            \\quad \\forall loc::tech \\in loc::techs_{store}, \\forall timestep \\in timesteps

    Where :math:`cluster(timestep)` is the cluster number in which the timestep
    is located.
    """
    cluster = backend_model.__calliope_model_data["data"]["timestep_cluster"][timestep]
    return (
        backend_model.storage[loc_tech, timestep]
        >= backend_model.storage_intra_cluster_min[loc_tech, cluster]
    )


def storage_inter_max_rule(backend_model, loc_tech, datestep):
    """
    When clustering days, to reduce the timeseries length, set maximum limit on
    the intra-cluster and inter-date stored energy.
    intra-cluster = all timesteps in a single cluster
    datesteps = all dates in the unclustered timeseries (each has a corresponding cluster)
    `Ref: DOI 10.1016/j.apenergy.2018.01.023 <https://doi.org/10.1016/j.apenergy.2018.01.023>`_

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{inter\\_cluster}}(loc::tech, datestep) +
            \\boldsymbol{storage_{intra\\_cluster, max}}(loc::tech, cluster(datestep))
            \\leq \\boldsymbol{storage_{cap}}(loc::tech) \\quad \\forall
            loc::tech \\in loc::techs_{store}, \\forall datestep \\in datesteps

    Where :math:`cluster(datestep)` is the cluster number in which the datestep
    is located.
    """
    cluster = backend_model.__calliope_model_data["data"]["lookup_datestep_cluster"][
        datestep
    ]
    return (
        backend_model.storage_inter_cluster[loc_tech, datestep]
        + backend_model.storage_intra_cluster_max[loc_tech, cluster]
        <= backend_model.storage_cap[loc_tech]
    )


def storage_inter_min_rule(backend_model, loc_tech, datestep):
    """
    When clustering days, to reduce the timeseries length, set minimum limit on
    the intra-cluster and inter-date stored energy.
    intra-cluster = all timesteps in a single cluster
    datesteps = all dates in the unclustered timeseries (each has a corresponding cluster)
    `Ref: DOI 10.1016/j.apenergy.2018.01.023 <https://doi.org/10.1016/j.apenergy.2018.01.023>`_

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{storage_{inter\\_cluster}}(loc::tech, datestep)
            \\times (1 - storage\\_loss(loc::tech, timestep))^{24} +
            \\boldsymbol{storage_{intra\\_cluster, min}}(loc::tech, cluster(datestep))
            \\geq 0 \\quad \\forall loc::tech \\in loc::techs_{store},
            \\forall datestep \\in datesteps

    Where :math:`cluster(datestep)` is the cluster number in which the datestep
    is located.
    """
    cluster = backend_model.__calliope_model_data["data"]["lookup_datestep_cluster"][
        datestep
    ]
    storage_loss = get_param(backend_model, "storage_loss", loc_tech)
    return (
        backend_model.storage_inter_cluster[loc_tech, datestep]
        * ((1 - storage_loss) ** 24)
        + backend_model.storage_intra_cluster_min[loc_tech, cluster]
        >= 0
    )



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

"""

from collections.abc import Mapping
from calliope.core.attrdict import AttrDict


class ObservedDict(dict):
    """
    Dictionary subclass which updates UpdateObserverDict when there is a change
    in the values assigned to keys in the dictionary.
    """

    def __init__(self, initial_dict, initial_yaml_string, on_changed=None):
        if initial_yaml_string is not None:
            initial_dict = AttrDict.from_yaml_string(initial_yaml_string).as_dict()

        super().__init__(initial_dict)

        self.on_changed = on_changed

        for k, v in initial_dict.items():
            if isinstance(v, dict):
                super().__setitem__(k, ObservedDict(v, None, on_changed=self.notify))
        self.notify()

    def __setitem__(self, key, value):
        if isinstance(value, dict):
            value = ObservedDict(value, None, on_changed=self.notify)
        super().__setitem__(key, value)
        self.notify()

    def update(self, other=None, **kwargs):
        if other is not None:
            for k, v in other.items() if isinstance(other, Mapping) else other:
                self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def notify(self, updated=None):
        if self.on_changed is not None:
            return self.on_changed(self)


class UpdateObserverDict(ObservedDict):
    """
    Dictionary subclass which observes a dictionary and updates an attribute of
    the model_data xarray dataset with a YAML string of that dictionary.
    This update takes place whenever a value is updated in the dictionary.

    Parameters
    ----------
    name : str
        The model_data attribute key to update.
    observer : xarray Dataset (e.g. calliope model_data)
        The Dataset whose attribute dictionary will contain the key `name`
        and value = a YAML string of the initial_dict.
    initial_dict : dict, optional, default = None
        An initial dictionary to copy for observing.
        One of initial_dict or initial_yaml_string must be defined.
    initial_yaml_string : str, optional, default = None
        A YAML string of an initial dictionary to copy for observing.
        One of initial_dict or initial_yaml_string must be defined.

    Returns
    -------
    Observed dictionary, which acts as a dictionary in every sense *except* that
    on changing or adding any key:value pairs in that dictionary, the YAML string
    stored at `observer.attrs[name]` will be updated to reflect the new Observed
    dicitonary.
    """

    def __init__(
        self,
        name,
        observer,
        initial_dict=None,
        initial_yaml_string=None,
        *args,
        **kwargs,
    ):
        self.observer = observer
        self.name = name

        check_input_args = [i is None for i in [initial_dict, initial_yaml_string]]
        if all(check_input_args) or not any(check_input_args):
            raise ValueError(
                "must supply one, and only one, of initial_dict or initial_yaml_string"
            )

        super().__init__(initial_dict, initial_yaml_string, *args, **kwargs)

    def notify(self, updated=None):
        temp_dict = {
            k: v
            for k, v in self.items()
            if (not isinstance(v, dict) and v is not None)
            or (isinstance(v, dict) and len(v.keys()) > 0)
        }
        self.observer.attrs[self.name] = AttrDict(temp_dict).to_yaml()



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

run_checks.py
~~~~~~~~~~~~~

Checks for model consistency and possible errors when preparing run in the backend.

"""
import numpy as np
import pandas as pd
import xarray as xr
from calliope.core.attrdict import AttrDict
from calliope.core.util.observed_dict import UpdateObserverDict


def check_operate_params(model_data):
    """
    if model mode = `operate`, check for clashes in capacity constraints.
    In this mode, all capacity constraints are set to parameters in the backend,
    so can easily lead to model infeasibility if not checked.

    Returns
    -------
    comments : AttrDict
        debug output
    warnings : list
        possible problems that do not prevent the model run
        from continuing
    errors : list
        serious issues that should raise a ModelError

    """
    defaults = UpdateObserverDict(
        initial_yaml_string=model_data.attrs["defaults"],
        name="defaults",
        observer=model_data,
    )
    run_config = UpdateObserverDict(
        initial_yaml_string=model_data.attrs["run_config"],
        name="run_config",
        observer=model_data,
    )

    warnings, errors = [], []
    comments = AttrDict()

    def _get_param(loc_tech, var):
        if _is_in(loc_tech, var) and not model_data[var].loc[loc_tech].isnull().any():
            param = model_data[var].loc[loc_tech].values
        else:
            param = defaults[var]
        return param

    def _is_in(loc_tech, set_or_var):
        try:
            model_data[set_or_var].loc[loc_tech]
            return True
        except (KeyError, AttributeError):
            return False

    def _set_inf_and_warn(loc_tech, var, warnings, warning_text):
        # inf will only work on float dtypes
        # astype doesn't retain attributes: https://github.com/pydata/xarray/issues/2049
        _attrs = model_data[var].attrs
        model_data[var] = model_data[var].astype(float)
        model_data[var].attrs = _attrs
        if np.isinf(model_data[var].loc[loc_tech].item()):
            return (np.inf, warnings)
        elif model_data[var].loc[loc_tech].isnull().item():
            var_name = model_data[var].loc[loc_tech] = np.inf
            return (var_name, warnings)
        else:
            var_name = model_data[var].loc[loc_tech] = np.inf
            warnings.append(warning_text)
            return var_name, warnings

    # Storage initial is carried over between iterations, so must be defined along with storage
    if (
        "loc_techs_store" in model_data.dims.keys()
        and "storage_initial" not in model_data.data_vars.keys()
    ):
        model_data["storage_initial"] = xr.DataArray(
            [0.0 for loc_tech in model_data.loc_techs_store.values],
            dims="loc_techs_store",
        )
        model_data["storage_initial"].attrs["is_result"] = 0.0
        warnings.append(
            "Initial stored energy not defined, set to zero for all "
            "loc::techs in loc_techs_store, for use in iterative optimisation"
        )

    # Operated units is carried over between iterations, so must be defined in a milp model
    if (
        "loc_techs_milp" in model_data.dims.keys()
        and "operated_units" not in model_data.data_vars.keys()
    ):
        model_data["operated_units"] = xr.DataArray(
            [0 for loc_tech in model_data.loc_techs_milp.values], dims="loc_techs_milp"
        )
        model_data["operated_units"].attrs["is_result"] = 1
        model_data["operated_units"].attrs["operate_param"] = 1
        warnings.append(
            "daily operated units not defined, set to zero for all "
            "loc::techs in loc_techs_milp, for use in iterative optimisation"
        )

    for loc_tech in model_data.loc_techs.values:
        energy_cap = model_data.energy_cap.loc[loc_tech].item()
        # Must have energy_cap defined for all relevant techs in the model
        if (np.isinf(energy_cap) or np.isnan(energy_cap)) and (
            _get_param(loc_tech, "energy_cap_min_use")
            or (
                _get_param(loc_tech, "force_resource")
                and _get_param(loc_tech, "resource_unit") == "energy_per_cap"
            )
        ):
            errors.append(
                "Operate mode: User must define a finite energy_cap (via "
                "energy_cap_equals or energy_cap_max) for {}".format(loc_tech)
            )

        elif _is_in(loc_tech, "loc_techs_finite_resource"):
            # Cannot have infinite resource area if linking resource and area (resource_unit = energy_per_area)
            if (
                _is_in(loc_tech, "loc_techs_area")
                and model_data.resource_unit.loc[loc_tech].item() == "energy_per_area"
            ):
                if _is_in(loc_tech, "resource_area"):
                    area = model_data.resource_area.loc[loc_tech].item()
                else:
                    area = None
                if pd.isnull(area) or np.isinf(area):
                    errors.append(
                        "Operate mode: User must define a finite resource_area "
                        "(via resource_area_equals or resource_area_max) for {}, "
                        "as available resource is linked to resource_area "
                        "(resource_unit = `energy_per_area`)".format(loc_tech)
                    )

            # force resource overrides capacity constraints, so set capacity constraints to infinity
            if _get_param(loc_tech, "force_resource"):

                if not _is_in(loc_tech, "loc_techs_store"):
                    # set resource_area to inf if the resource is linked to energy_cap using energy_per_cap
                    if (
                        model_data.resource_unit.loc[loc_tech].item()
                        == "energy_per_cap"
                    ):
                        if _is_in(loc_tech, "resource_area"):
                            resource_area, warnings = _set_inf_and_warn(
                                loc_tech,
                                "resource_area",
                                warnings,
                                "Resource area constraint removed from {} as "
                                "force_resource is applied and resource is linked "
                                "to energy flow using `energy_per_cap`".format(
                                    loc_tech
                                ),
                            )
                    # set energy_cap to inf if the resource is linked to resource_area using energy_per_area
                    elif (
                        model_data.resource_unit.loc[loc_tech].item()
                        == "energy_per_area"
                    ):
                        energy_cap, warnings = _set_inf_and_warn(
                            loc_tech,
                            "energy_cap",
                            warnings,
                            "Energy capacity constraint removed from {} as "
                            "force_resource is applied and resource is linked "
                            "to energy flow using `energy_per_area`".format(loc_tech),
                        )
                    # set both energy_cap and resource_area to inf if the resource is not linked to anything
                    elif model_data.resource_unit.loc[loc_tech].item() == "energy":
                        if _is_in(loc_tech, "resource_area"):
                            resource_area, warnings = _set_inf_and_warn(
                                loc_tech,
                                "resource_area",
                                warnings,
                                "Resource area constraint removed from {} as "
                                "force_resource is applied and resource is not linked "
                                "to energy flow (resource_unit = `energy`)".format(
                                    loc_tech
                                ),
                            )
                        energy_cap, warnings = _set_inf_and_warn(
                            loc_tech,
                            "energy_cap",
                            warnings,
                            "Energy capacity constraint removed from {} as "
                            "force_resource is applied and resource is not linked "
                            "to energy flow (resource_unit = `energy`)".format(
                                loc_tech
                            ),
                        )

                if _is_in(loc_tech, "resource_cap"):
                    resource_cap, warnings = _set_inf_and_warn(
                        loc_tech,
                        "resource_cap",
                        warnings,
                        "Resource capacity constraint removed from {} as "
                        "force_resource is applied".format(loc_tech),
                    )

        if _is_in(loc_tech, "loc_techs_store"):
            storage_cap = model_data.storage_cap.loc[loc_tech].item()
            if not pd.isnull(storage_cap) and not pd.isnull(energy_cap):
                if _get_param(loc_tech, "charge_rate") is not False:
                    charge_rate = model_data["charge_rate"].loc[loc_tech].item()
                    if storage_cap * charge_rate < energy_cap:
                        errors.append(
                            "fixed storage capacity * charge_rate is not larger "
                            "than fixed energy capacity for loc::tech {}".format(
                                loc_tech
                            )
                        )
                if _get_param(loc_tech, "energy_cap_per_storage_cap_max") is not False:
                    energy_cap_per_storage_cap_max = (
                        model_data["energy_cap_per_storage_cap_max"]
                        .loc[loc_tech]
                        .item()
                    )
                    if storage_cap * energy_cap_per_storage_cap_max < energy_cap:
                        errors.append(
                            "fixed storage capacity * energy_cap_per_storage_cap_max is not larger "
                            "than fixed energy capacity for loc::tech {}".format(
                                loc_tech
                            )
                        )
                if _get_param(loc_tech, "energy_cap_per_storage_cap_min") is not False:
                    energy_cap_per_storage_cap_min = (
                        model_data["energy_cap_per_storage_cap_min"]
                        .loc[loc_tech]
                        .item()
                    )
                    if storage_cap * energy_cap_per_storage_cap_min > energy_cap:
                        errors.append(
                            "fixed storage capacity * energy_cap_per_storage_cap_min is not smaller "
                            "than fixed energy capacity for loc::tech {}".format(
                                loc_tech
                            )
                        )
    # Must define a resource capacity to ensure the Pyomo param is created
    # for it. But we just create an array of infs, so the capacity has no effect
    if (
        "resource_cap" not in model_data.data_vars.keys()
        and "loc_techs_supply_plus" in model_data.dims.keys()
    ):
        model_data["resource_cap"] = xr.DataArray(
            [np.inf for i in model_data.loc_techs_supply_plus.values],
            dims="loc_techs_supply_plus",
        )
        model_data["resource_cap"].attrs["is_result"] = 1
        model_data["resource_cap"].attrs["operate_param"] = 1
        warnings.append(
            "Resource capacity constraint defined and set to infinity "
            "for all supply_plus techs"
        )

    window = run_config.get("operation", {}).get("window", None)
    horizon = run_config.get("operation", {}).get("horizon", None)
    if not window or not horizon:
        errors.append(
            "Operational mode requires a timestep window and horizon to be "
            "defined under run.operation"
        )
    elif horizon < window:
        errors.append(
            "Iteration horizon must be larger than iteration window, "
            "for operational mode"
        )

    # Cyclic storage isn't really valid in operate mode, so we ignore it, using
    # initial_storage instead (allowing us to pass storage between operation windows)
    if run_config.get("cyclic_storage", True):
        warnings.append(
            "Storage cannot be cyclic in operate run mode, setting "
            "`run.cyclic_storage` to False for this run"
        )
        run_config["cyclic_storage"] = False

    if "group_demand_share_per_timestep_decision" in model_data.data_vars:
        warnings.append(
            "`demand_share_per_timestep_decision` group constraints cannot be "
            "used in operate mode, so will not be built."
        )
        del model_data["group_demand_share_per_timestep_decision"]

    return comments, warnings, errors



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

preprocess_sets.py
~~~~~~~~~~~~~~~~~~

Sets & sub-sets defined to reduce size of decision variables & constraints.

The markers preceded with `###` are used to auto-include the set description
in the documentation

###PART TO INCLUDE IN DOCUMENTATION STARTS HERE###

Main sets
=========

* timesteps
* techs
* techs_non_transmission
* techs_transmission
* locs
* costs
* resources
* carriers
* carrier_tiers
* techlists

Location-technology subsets
===========================

* loc_techs

Technology groups

* loc_techs_storage
* loc_techs_transmission
* loc_techs_demand
* loc_techs_supply
* loc_techs_supply_plus
* loc_techs_conversion
* loc_techs_conversion_plus

Subsets based on active constraints

* loc_techs_area
* loc_techs_store
* loc_techs_finite_resource
* loc_techs_finite_resource_supply
* loc_techs_finite_resource_demand
* loc_techs_finite_resource_supply_plus
* loc_techs_ramping
* loc_techs_export
* loc_techs_purchase
* loc_techs_milp
* loc_techs_cost
* loc_techs_investment_cost
* loc_techs_om_cost
* loc_techs_out_2
* loc_techs_out_3
* loc_techs_in_2
* loc_techs_in_3


Subsets that include carrier

* loc_tech_carriers_prod
* loc_tech_carriers_con
* loc_tech_carriers_conversion_plus
* loc_tech_carriers_export
* loc_carriers

###PART TO INCLUDE IN DOCUMENTATION ENDS HERE###

"""

from itertools import product

import numpy as np

from calliope.core.attrdict import AttrDict
from calliope.preprocess.util import (
    get_all_carriers,
    split_loc_techs_transmission,
    concat_iterable,
    flatten_list,
)


def generate_simple_sets(model_run):
    """
    Generate basic sets for a given pre-processed ``model_run``.

    Parameters
    ----------
    model_run : AttrDict

    """
    sets = AttrDict()

    flat_techs = model_run.techs.as_dict(flat=True)
    flat_locations = model_run.locations.as_dict(flat=True)

    sets.resources = set(
        flatten_list(v for k, v in flat_techs.items() if ".carrier" in k)
    )

    sets.carriers = sets.resources - set(["resource"])

    sets.carrier_tiers = set(
        key.split(".carrier_")[1] for key in flat_techs.keys() if ".carrier_" in key
    )

    sets.costs = set(
        k.split("costs.")[-1].split(".")[0]
        for k in flat_locations.keys()
        if ".costs." in k
    )

    sets.locs = set(model_run.locations.keys())

    sets.techs_non_transmission = set()
    tech_groups = [
        "demand",
        "supply",
        "supply_plus",
        "conversion",
        "conversion_plus",
        "storage",
    ]
    for tech_group in tech_groups:
        sets["techs_{}".format(tech_group)] = set(
            k for k, v in model_run.techs.items() if v.inheritance[-1] == tech_group
        )
        sets.techs_non_transmission.update(sets["techs_{}".format(tech_group)])

    sets.techs_transmission_names = set(
        k for k, v in model_run.techs.items() if v.inheritance[-1] == "transmission"
    )

    # This builds the "tech:loc" expansion of transmission technologies
    techs_transmission = set()
    for loc_name, loc_config in model_run.locations.items():
        for link_name, link_config in loc_config.get("links", {}).items():
            for tech_name in link_config.techs:
                techs_transmission.add("{}:{}".format(tech_name, link_name))
    sets.techs_transmission = techs_transmission

    sets.techs = sets.techs_non_transmission | sets.techs_transmission_names

    # this extracts location coordinate information
    coordinates = set(
        k.split(".")[-1] for k in flat_locations.keys() if ".coordinates." in k
    )

    if coordinates:
        sets.coordinates = coordinates

    # `timesteps` set is built from the results of timeseries_data processing
    sets.timesteps = list(model_run.timesteps.astype(str))
    model_run.del_key("timesteps")

    # `techlists` are strings with comma-separated techs used for grouping in
    # some model-wide constraints
    sets.techlists = set()
    for k in model_run.model.get_key("group_share", {}).keys():
        sets.techlists.add(k)

    return sets


def generate_loc_tech_sets(model_run, simple_sets):
    """
    Generate loc-tech sets for a given pre-processed ``model_run``

    Parameters
    ----------
    model_run : AttrDict
    simple_sets : AttrDict
        Simple sets returned by ``generate_simple_sets(model_run)``.

    """
    sets = AttrDict()

    ##
    # First deal with transmission techs, which can show up only in
    # loc_techs_transmission, loc_techs_milp, and loc_techs_purchase
    ##

    # All `tech:loc` expanded transmission technologies
    sets.loc_techs_transmission = set(
        concat_iterable(
            [
                (i, u, j)
                for i, j, u in product(  # (loc, loc, tech) product
                    simple_sets.locs,
                    simple_sets.locs,
                    simple_sets.techs_transmission_names,
                )
                if model_run.get_key(
                    "locations.{}.links.{}.techs.{}".format(i, j, u), None
                )
            ],
            ["::", ":"],
        )
    )

    # A dict of transmission tech config objects
    # to make parsing for set membership easier
    loc_techs_transmission_config = {
        k: model_run.get_key(
            "locations.{loc_from}.links.{loc_to}.techs.{tech}".format(
                **split_loc_techs_transmission(k)
            )
        )
        for k in sets.loc_techs_transmission
    }

    ##
    # Now deal with the rest of the techs and other sets
    ##

    # Only loc-tech combinations that actually exist
    sets.loc_techs_non_transmission = set(
        concat_iterable(
            [
                (l, t)
                for l, t in product(
                    simple_sets.locs, simple_sets.techs_non_transmission
                )
                if model_run.get_key("locations.{}.techs.{}".format(l, t), None)
            ],
            ["::"],
        )
    )

    sets.loc_techs = sets.loc_techs_non_transmission | sets.loc_techs_transmission

    # A dict of non-transmission tech config objects
    # to make parsing for set membership easier
    loc_techs_config = {
        k: model_run.get_key("locations.{}.techs.{}".format(*k.split("::")))
        for k in sets.loc_techs_non_transmission
    }

    loc_techs_all_config = {**loc_techs_config, **loc_techs_transmission_config}

    ##
    # Sets based on membership in abstract base technology groups
    ##

    for group in [
        "storage",
        "demand",
        "supply",
        "supply_plus",
        "conversion",
        "conversion_plus",
    ]:
        tech_set = set(
            k
            for k in sets.loc_techs_non_transmission
            if model_run.techs[k.split("::")[1]].inheritance[-1] == group
        )
        sets["loc_techs_{}".format(group)] = tech_set

    sets.loc_techs_non_conversion = (
        set(
            k
            for k in sets.loc_techs_non_transmission
            if k not in sets.loc_techs_conversion
            and k not in sets.loc_techs_conversion_plus
        )
        | sets.loc_techs_transmission
    )

    # Techs that introduce energy into the system
    sets.loc_techs_supply_all = sets.loc_techs_supply | sets.loc_techs_supply_plus

    # Techs that change the energy carrier in the system
    sets.loc_techs_conversion_all = (
        sets.loc_techs_conversion | sets.loc_techs_conversion_plus
    )
    # All techs that can be used to generate a carrier (not just store or move it)
    sets.loc_techs_supply_conversion_all = (
        sets.loc_techs_supply_all | sets.loc_techs_conversion_all
    )

    ##
    # Sets based on specific constraints being active
    ##

    # Technologies that specify resource_area constraints
    sets.loc_techs_area = set(
        k
        for k in sets.loc_techs_non_transmission
        if (
            any("resource_area" in i for i in loc_techs_config[k].keys_nested())
            or loc_techs_config[k].constraints.get("resource_unit", "energy")
            == "energy_per_area"
        )
    )

    # Technologies that define storage, which can include `supply_plus`
    # and `storage` groups.
    sets.loc_techs_store = (
        set(
            k
            for k in sets.loc_techs_supply_plus
            if any(
                "storage_" in i for i in loc_techs_config[k].constraints.keys_nested()
            )
        )
        | sets.loc_techs_storage
    )

    # technologies that specify a finite resource
    sets.loc_techs_finite_resource = set(
        k
        for k in sets.loc_techs_non_transmission
        if loc_techs_config[k].constraints.get("resource")
        and not (loc_techs_config[k].constraints.get("resource") in ["inf", np.inf])
    )

    # `supply` technologies that specify a finite resource
    sets.loc_techs_finite_resource_supply = sets.loc_techs_finite_resource.intersection(
        sets.loc_techs_supply
    )

    # `demand` technologies that specify a finite resource
    sets.loc_techs_finite_resource_demand = sets.loc_techs_finite_resource.intersection(
        sets.loc_techs_demand
    )

    # `supply_plus` technologies that specify a finite resource
    sets.loc_techs_finite_resource_supply_plus = (
        sets.loc_techs_finite_resource.intersection(sets.loc_techs_supply_plus)
    )

    # Technologies that define ramping constraints
    sets.loc_techs_ramping = set(
        k
        for k in sets.loc_techs_non_transmission
        if "energy_ramping" in loc_techs_config[k].constraints
    )

    # Technologies that allow export
    sets.loc_techs_export = set(
        k
        for k in sets.loc_techs_non_transmission
        if "export_carrier" in loc_techs_config[k].constraints
    )

    # Technologies that allow purchasing discrete units
    # NB: includes transmission techs!
    loc_techs_purchase = set(
        k
        for k in sets.loc_techs_non_transmission
        if any(
            ".purchase" in i
            for i in loc_techs_config[k].get("costs", AttrDict()).keys_nested()
        )
        and not any(
            "units_" in i
            for i in loc_techs_config[k].get("constraints", AttrDict()).keys_nested()
        )
    )

    transmission_purchase = set(
        k
        for k in sets.loc_techs_transmission
        if any(
            ".purchase" in i
            for i in loc_techs_transmission_config[k]
            .get("costs", AttrDict())
            .keys_nested()
        )
        and not any(
            "units_" in i
            for i in loc_techs_transmission_config[k]
            .get("constraints", AttrDict())
            .keys_nested()
        )
    )

    sets.loc_techs_purchase = loc_techs_purchase | transmission_purchase

    # Technologies with MILP constraints
    loc_techs_milp = set(
        k
        for k in sets.loc_techs_non_transmission
        if any("units_" in i for i in loc_techs_config[k].constraints.keys_nested())
    )

    transmission_milp = set(
        k
        for k in sets.loc_techs_transmission
        if any(
            "units_" in i
            for i in loc_techs_transmission_config[k].constraints.keys_nested()
        )
    )

    sets.loc_techs_milp = loc_techs_milp | transmission_milp

    # Technologies with forced asynchronous production/consumption of energy
    loc_techs_storage_asynchronous_prod_con = set(
        k
        for k in sets.loc_techs_store
        if "force_asynchronous_prod_con"
        in loc_techs_config[k].constraints.keys_nested()
    )

    loc_techs_transmission_asynchronous_prod_con = set(
        k
        for k in sets.loc_techs_transmission
        if "force_asynchronous_prod_con"
        in loc_techs_transmission_config[k].constraints.keys_nested()
    )
    sets.loc_techs_asynchronous_prod_con = (
        loc_techs_storage_asynchronous_prod_con
        | loc_techs_transmission_asynchronous_prod_con
    )

    ##
    # Sets based on specific costs being active
    # NB includes transmission techs
    ##

    loc_techs_costs = set(
        k
        for k in sets.loc_techs_non_transmission
        if any("costs" in i for i in loc_techs_config[k].keys())
    )

    loc_techs_transmission_costs = set(
        k
        for k in sets.loc_techs_transmission
        if any("costs" in i for i in loc_techs_transmission_config[k].keys())
    )

    # Any capacity or fixed annual costs
    loc_techs_investment_costs = set(
        k
        for k in loc_techs_costs
        if any(
            "_cap" in i or ".purchase" in i or "_area" in i
            for i in loc_techs_config[k].costs.keys_nested()
        )
    )
    loc_techs_transmission_investment_costs = set(
        k
        for k in loc_techs_transmission_costs
        if any(
            "_cap" in i or ".purchase" in i or "_area" in i
            for i in loc_techs_transmission_config[k].costs.keys_nested()
        )
    )

    # Any operation and maintenance
    loc_techs_om_costs = set(
        k
        for k in loc_techs_costs
        if any(
            "om_" in i or "export" in i for i in loc_techs_config[k].costs.keys_nested()
        )
    )
    loc_techs_transmission_om_costs = set(
        k
        for k in loc_techs_transmission_costs
        if any("om_" in i for i in loc_techs_transmission_config[k].costs.keys_nested())
    )

    # Any export costs
    sets.loc_techs_costs_export = set(
        k
        for k in loc_techs_costs
        if any("export" in i for i in loc_techs_config[k].costs.keys_nested())
    )

    sets.loc_techs_cost = loc_techs_costs | loc_techs_transmission_costs
    sets.loc_techs_investment_cost = (
        loc_techs_investment_costs | loc_techs_transmission_investment_costs
    )
    sets.loc_techs_om_cost = loc_techs_om_costs | loc_techs_transmission_om_costs

    ##
    # Subsets of costs for different abstract base technologies
    ##

    sets.loc_techs_om_cost_conversion = loc_techs_om_costs.intersection(
        sets.loc_techs_conversion
    )
    sets.loc_techs_om_cost_conversion_plus = loc_techs_om_costs.intersection(
        sets.loc_techs_conversion_plus
    )
    sets.loc_techs_om_cost_supply = loc_techs_om_costs.intersection(
        sets.loc_techs_supply
    )
    sets.loc_techs_om_cost_supply_plus = loc_techs_om_costs.intersection(
        sets.loc_techs_supply_plus
    )

    ##
    # Subsets of `conversion_plus` technologies
    ##

    # `conversion_plus` technologies with secondary carrier(s) out
    sets.loc_techs_out_2 = set(
        k
        for k in sets.loc_techs_conversion_plus
        if "carrier_out_2" in model_run.techs[k.split("::")[1].split(":")[0]].essentials
    )

    # `conversion_plus` technologies with tertiary carrier(s) out
    sets.loc_techs_out_3 = set(
        k
        for k in sets.loc_techs_conversion_plus
        if "carrier_out_3" in model_run.techs[k.split("::")[1].split(":")[0]].essentials
    )

    # `conversion_plus` technologies with secondary carrier(s) in
    sets.loc_techs_in_2 = set(
        k
        for k in sets.loc_techs_conversion_plus
        if "carrier_in_2" in model_run.techs[k.split("::")[1].split(":")[0]].essentials
    )

    # `conversion_plus` technologies with tertiary carrier(s) in
    sets.loc_techs_in_3 = set(
        k
        for k in sets.loc_techs_conversion_plus
        if "carrier_in_3" in model_run.techs[k.split("::")[1].split(":")[0]].essentials
    )

    ##
    # `loc_tech_carrier` sets
    ##

    # loc_tech_carriers for all technologies that have energy_prod=True
    sets.loc_tech_carriers_prod = set(
        "{}::{}".format(k, carrier)
        for k in sets.loc_techs
        if loc_techs_all_config[k].constraints.get_key("energy_prod", False)
        for carrier in get_all_carriers(
            model_run.techs[k.split("::")[1].split(":")[0]].essentials, direction="out"
        )
    )

    # loc_tech_carriers for all technologies that have energy_con=True
    sets.loc_tech_carriers_con = set(
        "{}::{}".format(k, carrier)
        for k in sets.loc_techs
        if loc_techs_all_config[k].constraints.get_key("energy_con", False)
        for carrier in get_all_carriers(
            model_run.techs[k.split("::")[1].split(":")[0]].essentials, direction="in"
        )
    )

    # loc_tech_carriers for all supply technologies
    sets.loc_tech_carriers_supply_all = set(
        "{}::{}".format(k, carrier)
        for k in sets.loc_techs_supply_all
        for carrier in get_all_carriers(
            model_run.techs[k.split("::")[1].split(":")[0]].essentials, direction="out"
        )
    )

    # loc_tech_carriers for all conversion technologies
    sets.loc_tech_carriers_conversion_all = set(
        "{}::{}".format(k, carrier)
        for k in sets.loc_techs_conversion_all
        for carrier in get_all_carriers(
            model_run.techs[k.split("::")[1].split(":")[0]].essentials, direction="out"
        )
    )

    # loc_tech_carriers for all supply and conversion technologies
    sets.loc_tech_carriers_supply_conversion_all = (
        sets.loc_tech_carriers_supply_all | sets.loc_tech_carriers_conversion_all
    )
    # loc_tech_carriers for all demand technologies
    sets.loc_tech_carriers_demand = set(
        "{}::{}".format(k, carrier)
        for k in sets.loc_techs_demand
        for carrier in get_all_carriers(
            model_run.techs[k.split("::")[1].split(":")[0]].essentials, direction="in"
        )
    )

    # loc_tech_carriers for all technologies that have export
    sets.loc_tech_carriers_export = set(
        "{}::{}".format(k, loc_techs_all_config[k].constraints.export_carrier)
        for k in sets.loc_techs
        if loc_techs_all_config[k].constraints.get_key("export_carrier", False)
    )

    # loc_tech_carriers for `conversion_plus` technologies
    sets.loc_tech_carriers_conversion_plus = set(
        k
        for k in sets.loc_tech_carriers_con | sets.loc_tech_carriers_prod
        if k.rsplit("::", 1)[0] in sets.loc_techs_conversion_plus
    )

    # loc_carrier combinations that exist with either a con or prod tech
    sets.loc_carriers = set(
        "{0}::{2}".format(*k.split("::"))
        for k in sets.loc_tech_carriers_prod | sets.loc_tech_carriers_con
    )

    return sets
